#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

from jgdv.structs.locator.errors import DirAbsent, LocationExpansionError, LocationError
from jgdv.structs.locator import JGDVLocator, Location
from jgdv.structs.locator.locator import _LocatorGlobal
from jgdv.structs.dkey import DKey, NonDKey

logging = logmod.root

match JGDVLocator.Current:
    case None:
        initial_loc = JGDVLocator(pl.Path.cwd())
    case x:
        initial_loc = x

@pytest.fixture(scope="function")
def simple() -> JGDVLocator:
    return JGDVLocator(pl.Path.cwd())

@pytest.fixture(scope="function")
def wrap_locs():
    logging.debug("Activating temp locs")
    with JGDVLocator(pl.Path.cwd()) as temp:
        yield temp
##--|

class TestLocator:

    def test_initial(self):
        simple = JGDVLocator(pl.Path.cwd())
        assert(isinstance(simple, JGDVLocator))
        assert(not bool(simple._data))

    def test_update(self, simple):
        assert(not bool(simple._data))
        simple.update({"blah": "file::>bloo"})
        assert(bool(simple._data))
        assert("blah" in simple)

    def test_data_stored_as_locations(self, simple):
        assert(not bool(simple._data))
        simple.update({"blah": "file::>bloo", "aweg":"aweg/abloo"})
        assert(bool(simple._data))
        for x in  simple._data.values():
            assert(isinstance(x, Location))

    def test_registered(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "file::>blah"})
        assert(bool(simple._data))
        simple.registered("a")

    def test_registered_fail(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "file::>blah"})
        assert(bool(simple._data))

        with pytest.raises(DirAbsent):
            simple.registered("b")

    def test_update_conflict(self, simple):
        simple.update({"blah": "dir::>bloo"})
        with pytest.raises(LocationError):
            simple.update({"blah": "dir::>blah"})

    def test_update_non_strict(self, simple):
        simple.update({"blah": "dir::>bloo"})
        simple.update({"blah": "dir::>bloo"}, strict=False)

    def test_update_overwrite(self, simple):
        locstr = "dir::>aweg"
        simple = JGDVLocator(pl.Path.cwd())
        simple.update({"blah": "dirr::bloo"})
        simple.update({"blah": "dir::>aweg"}, strict=False)
        assert("blah" in simple)
        assert(simple._data["blah"] == locstr)
        assert(simple.access('blah') == locstr)

    def test_empty_repr(self, simple):
        repr_str = repr(simple)
        assert(repr_str == f"<JGDVLocator (1) : {str(pl.Path.cwd())} : ()>")

    def test_non_empty_repr(self, simple):
        simple.update({"a": "dir::>blah", "b": "dir::>aweg", "awegewag": "dir::>jkwejgio"})
        repr_str = repr(simple)
        assert(repr_str == f"<JGDVLocator (1) : {str(pl.Path.cwd())} : (a, b, awegewag)>")

    def test_clear(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "dir::>blah"})
        assert("a" in simple)
        simple.clear()
        assert("a" not in simple)

    @pytest.mark.xfail
    def test_metacheck(self, simple):
        assert(False)

class TestLocatorExpansion:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_expansion_attr(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        match wrap_locs.todo_bib:
            case Location() as x:
                assert(x.path == pl.Path("~/github/bibliography/in_progress/todo.bib"))
                assert(True)
            case x:
                 assert(False), x

    def test_expansion_item(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        target = pl.Path("~/github/bibliography/in_progress/todo.bib").expanduser().resolve()
        match wrap_locs['{todo_bib}']:
            case pl.Path() as x:
                assert(x == target)
            case x:
                 assert(False), x

    def test_expansion_no_key(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        target = pl.Path("todo_bib").resolve()
        match wrap_locs['todo_bib']:
            case pl.Path() as x:
                assert(x == target)
            case x:
                 assert(False), x

class TestLocatorUtils:

    def test_normalize(self, simple):
        a_path = pl.Path("a/b/c")
        expected = a_path.absolute()
        result = simple.normalize(a_path)
        assert(result == expected)

    def test_normalize_tilde(self, simple):
        result = simple.normalize(pl.Path("~/blah"))
        assert(result.is_absolute())
        assert(result == pl.Path("~/blah").expanduser())

    def test_normalize_absolute(self, simple):
        result = simple.normalize(pl.Path("/blah"))
        assert(result.is_absolute())
        assert(result == pl.Path("/blah"))

    def test_normalize_relative(self, simple):
        result = simple.normalize(pl.Path("blah"))
        assert(result.is_absolute())
        assert(result == (pl.Path.cwd() / "blah").absolute())

    def test_normalize_relative_with_different_cwd(self):
        simple = JGDVLocator(pl.Path("~/desktop/"))
        result = simple.normalize(pl.Path("blah"))
        assert(result.is_absolute())
        assert(result == (pl.Path("~/desktop/") / "blah").expanduser().absolute())

class TestLocatorGlobal:

    def test_sanity(self, simple):
        assert(True is not False)

    def test_global_Current(self, simple):
        locs = JGDVLocator(pl.Path.cwd())
        assert(isinstance(JGDVLocator.Current, JGDVLocator))

    def test_ctx_manager_basic(self, simple):
        assert(JGDVLocator.Current is initial_loc)
        with JGDVLocator.Current() as locs2:
            assert(JGDVLocator.Current is locs2)

        assert(JGDVLocator.Current is initial_loc)

    def test_ctx_manager_cwd_change(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "dir::>blah"})
        assert(bool(simple._data))
        assert(simple.root == pl.Path.cwd())
        target = pl.Path("~/Desktop").expanduser().resolve()
        with simple(pl.Path("~/Desktop")) as ctx:
            assert(ctx.root == target)

    def test_stacklen(self, simple):
        assert(_LocatorGlobal.stacklen() == 1)
        locs  = JGDVLocator(pl.Path.cwd())
        assert(_LocatorGlobal.stacklen() == 1)
        with locs() as locs2:
            assert(_LocatorGlobal.stacklen() == 2)
            with locs2() as locs3:
                assert(_LocatorGlobal.stacklen() == 3)

            assert(_LocatorGlobal.stacklen() == 2)

        assert(_LocatorGlobal.stacklen() == 1)
        assert(JGDVLocator.Current is initial_loc)



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import abc
import atexit#  for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import types
import warnings
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from time import sleep
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generator,
                    Generic, Iterable, Iterator, Mapping, Match,
                    MutableMapping, Protocol, Sequence, Tuple, TypeAlias,
                    TypeGuard, TypeVar, cast, final, overload,
                    runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv.structs.chainguard._base import GuardBase
from jgdv.structs.chainguard.errors import GuardedAccessError
from jgdv.structs.chainguard.proxies.failure import GuardFailureProxy
from jgdv.structs.chainguard import ChainGuard
from jgdv.structs.chainguard.mixins.reporter_m import DefaultedReporter_m

# ##-- end 1st party imports

logging = logmod.root

class TestDefaultedReporter:

    @pytest.fixture(scope="function")
    def setup(self):
        pass

    @pytest.fixture(scope="function")
    def cleanup(self):
        pass

    def test_sanity(self):
        assert(True is True)

    def test_proxied_report_empty(self, mocker):
        mocker.patch.object(DefaultedReporter_m, "_defaulted", set())
        base     = ChainGuard({"test": { "blah": {"bloo": "final", "aweg": "joijo"}}})
        assert(ChainGuard.report_defaulted() == [])

    def test_proxied_report_no_existing_values(self, mocker):
        mocker.patch.object(DefaultedReporter_m, "_defaulted", set())
        base     = ChainGuard({"test": { "blah": {"bloo": "final", "aweg": "joijo"}}})
        base.test.blah.bloo
        base.test.blah.aweg
        assert(ChainGuard.report_defaulted() == [])

    def test_proxied_report_missing_values(self, mocker):
        mocker.patch.object(DefaultedReporter_m, "_defaulted", set())
        base              = ChainGuard({"test": { "blah": {"bloo": "final", "aweg": "joijo"}}})
        base.on_fail(False).this.doesnt.exist()
        base.on_fail(False).test.blah.other()

        defaulted = ChainGuard.report_defaulted()
        assert("<root>.this.doesnt.exist = false # <Any>" in defaulted)
        assert("<root>.test.blah.other = false # <Any>" in defaulted)

    def test_proxied_report_missing_typed_values(self, mocker):
        mocker.patch.object(DefaultedReporter_m, "_defaulted", set())
        base     = ChainGuard({"test": { "blah": {"bloo": "final", "aweg": "joijo"}}})
        base.on_fail("aValue", str).this.doesnt.exist()
        base.on_fail(2, int).test.blah.other()

        defaulted = ChainGuard.report_defaulted()
        assert("<root>.this.doesnt.exist = 'aValue' # <str>" in defaulted)
        assert("<root>.test.blah.other = 2 # <int>" in defaulted)

    @pytest.mark.skip("not implemented")
    def test_proxied_report_no_duplicates(self):
        raise NotImplementedError()



#!/usr/bin/env python3
"""
Provdes the Main ArgParser_p Protocol,
and the ParseMachineBase StateMachine.

ParseMachineBase descibes the state progression to parse arguments,
while jgdv.cli.arg_parser.CLIParser adds the specific logic to states and transitions
"""
# ruff: noqa: F401
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from collections import ChainMap
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from statemachine import State, StateMachine
from statemachine.states import States

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Maybe
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from .param_spec import ParamSpec

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
from ._interface import ParamStruct_p, ParamSource_p, ArgParser_p
# isort: on
# ##-- end types

 #-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class ParseMachineBase(StateMachine):
    """ Base Implementaiton of an FSM for running a CLI arg parse.
    Subclass and init with a default ArgParser_p that has bound callback for events
    """

    # States
    Start         = State(initial=True)
    Prepare       = State()
    Head          = State()
    CheckForHelp  = State()
    Cmd           = State()
    SubCmd        = State()
    Extra         = State()
    Cleanup       = State()
    ReadyToReport = State()
    Failed        = State()
    End           = State(final=True)

    # Event Transitions
    setup = (Prepare.to(Failed,            cond="_parse_fail_cond")
             | Prepare.to(ReadyToReport,   cond="_has_no_more_args_cond")
             | Start.to(Prepare,           after="setup")
             | Prepare.to(CheckForHelp,    after="setup")
             | CheckForHelp.to(Head)
             )

    parse = (Failed.from_(Start, Prepare, CheckForHelp, Head, Cmd, SubCmd, cond="_parse_fail_cond")
             | ReadyToReport.from_(Head, Cmd, SubCmd, Extra, cond="_has_no_more_args_cond")
             | Head.to(Cmd,      after="parse")
             | Cmd.to(SubCmd,    after="parse")
             | SubCmd.to(Extra,  after="parse")
             | Extra.to(Failed)
             )

    finish  = (End.from_(Cleanup)
               | ReadyToReport.to(Cleanup, after="finish")
               | Failed.to(Cleanup, after="finish")
               )

    def __init__(self, *, parser:Maybe[ArgParser_p]=None):
        assert(parser is not None)
        super().__init__(parser)
        self.count = 0
        self.max_attempts = 20

    def on_exit_state(self):
        self.count += 1
        if self.max_attempts < self.count:
            raise StopIteration

    def __call__(self, args:list[str], *, head_specs:list[ParamSpec], cmds:list[ParamSource_p], subcmds:list[tuple[str, ParamSource_p]]) -> Maybe[dict]:
        raise NotImplementedError()



 #!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import weakref
from collections import defaultdict
from uuid import UUID, uuid1

# ##-- end stdlib imports

from . import _interface as API  # noqa: N812
# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
from jgdv import Rx  # noqa: TC001

if TYPE_CHECKING:
    import pathlib as pl
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class TagFile(BaseModel):
    """ A Basic TagFile holds the counts for each tag use

    Tag file format is single lines of:
    ^{tag} {sep} {count}$

    cls.read can be used to change the {sep}

    # TODO use a collections.Counter
    """

    counts       : dict[str, int]        = defaultdict(lambda: 0)
    sep          : str                   = API.SEP
    ext          : str                   = API.TAG_EXT
    norm_replace : str                   = API.NORM_REPLACE
    norm_regex   : Rx                    = API.TAG_NORM
    comment      : str                   = API.COMMENT

    @classmethod
    def read(cls, fpath:pl.Path, **kwargs:dict) -> TagFile:
        obj = cls(**{x:y for x,y in kwargs.items() if y is not None})
        for i, line in enumerate(fpath.read_text().split("\n")):
            try:
                obj.update(line)
            except Exception as err:
                logging.warning("Failure Tag Read: (l:%s) : %s : %s : (file: %s)", i, err, line, fpath)
                raise

        return obj

    @field_validator("norm_regex", mode="before")
    def _validate_regex(cls, val:str|Rx) -> Rx:  # noqa: N805
        match val:
            case str():
                return re.compile(val)
            case re.Pattern():
                return val
            case _:
                msg = "Bad norm_regex provided"
                raise TypeError(msg)

    @field_validator("counts", mode="before")
    def _validate_counts(cls, val:dict) -> dict:  # noqa: N805
        counts : dict[str, int] = defaultdict(lambda: 0)
        match val:
            case dict():
                counts.update(val)
            case x:
                raise TypeError(type(x))
        return counts

    @model_validator(mode="after")
    def _normalize_counts(self) -> Self:
        orig = self.counts
        self.counts = defaultdict(lambda: 0)
        self.counts.update({self.norm_tag(x):y for x,y in orig.items()})
        return self

    def __iter__(self) -> Iterator:
        return iter(self.counts)

    def __str__(self) -> str:
        """
        Export the counts, 1 entry per line, as:
        `key` : `value`
        """
        all_lines = []
        for key in sorted(self.counts.keys(), key=lambda x: x.lower()):
            if not bool(self.counts[key]):
                continue
            all_lines.append(self.sep.join([key, str(self.counts[key])]))
        return "\n".join(all_lines)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __iadd__(self, values:TagFile|str|dict|set) -> Self:
        """  merge tags, updating their counts as well. """
        return self.update(values)

    def __len__(self) -> int:
        return len(self.counts)

    def __contains__(self, value:str) -> bool:
        return self.norm_tag(value) in self.counts

    def _inc(self, key:str, *, amnt:int=1) -> Maybe[str]:
        """ Increment the count of a key,
        if the normalized tag is empty, returns nothing
        """
        norm_key = self.norm_tag(key)
        if not bool(norm_key):
            return None
        self.counts[norm_key] += amnt
        return norm_key

    def update(self, *values:str|tuple[str,int|str]|TagFile|set|dict) -> Self:
        for val in values:
            match val:
                case None | "":
                    continue
                case str() if val.startswith(self.comment):
                    continue
                case str() if self.sep in val:
                    self.update(*(x.strip() for x in val.split(self.sep)))
                case str():
                    self._inc(val)
                case list() | set():
                    self.update(*val)
                case dict():
                    self.update(*val.items())
                case (str() as key, int()|str() as counts):
                    self._inc(key, amnt=int(counts))
                case TagFile():
                    self.update(*val.counts.items())
        else:
            return self

    def to_set(self) -> set[str]:
        return set(self.counts.keys())

    def get_count(self, tag:str) -> int:
        return self.counts[self.norm_tag(tag)]

    def norm_tag(self, tag:str) -> str:
        stripped = tag.strip()
        return self.norm_regex.sub(self.norm_replace, stripped).strip()



#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
default_log_colours : Final[dict[int, tuple[str, str]]] = {
    logmod.DEBUG    : ("fg", "grey"),
    logmod.INFO     : ("fg", "blue"),
    logmod.WARNING  : ("fg", "yellow"),
    logmod.ERROR    : ("fg", "red"),
    logmod.CRITICAL : ("fg", "red"),
}

default_colour_mapping : Final[dict[str, tuple[str,str]]] = {
    "blue"           : ("fg", "blue"),
    "cyan"           : ("fg", "cyan"),
    "green"          : ("fg", "green"),
    "magenta"        : ("fg", "magenta"),
    "red"            : ("fg", "red"),
    "yellow"         : ("fg", "yellow"),
    "bg_blue"        : ("bg", "blue"),
    "bg_cyan"        : ("bg", "cyan"),
    "bg_green"       : ("bg", "green"),
    "bg_magenta"     : ("bg", "magenta"),
    "bg_red"         : ("bg", "red"),
    "bg_yellow"      : ("bg", "yellow"),
    "bold"           : ("ef", "bold"),
    "underline"      : ("ef", "u"),
    "italic"         : ("ef", "italic"),
    "RESET"          : ("rs", "all"),
}
# Body:



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

import sys
from .. import JGDVError

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:
class TestBaseError:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133


    def test_simple(self):
        assert(True is not False) # noqa: PLR0133

    def test_traceback_slice(self):
        with pytest.raises(JGDVError) as ctx:
            raise JGDVError("blah")[:]

        target_line = sys._getframe().f_lineno - 2
        assert(ctx.tb.tb_lineno == target_line)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from jgdv._abstract.pre_processable import PreProcessor_p
from jgdv.structs.strang import _interface as StrangAPI  # noqa: N812
from jgdv.structs.strang.processor import StrangBasicProcessor
from . import _interface as API # noqa: N812

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from jgdv._abstract.pre_processable import InstanceData, PostInstanceData
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv._abstract.pre_processable import PreProcessResult
    from jgdv import Maybe, Ctor

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class LocationProcessor[T:API.Location_p](StrangBasicProcessor):

    @override
    def pre_process(self, cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> PreProcessResult[T]: # type: ignore[override]
        x             : Any
        y             : Any
        default_mark  : StrangAPI.StrangMarkAbstract_e
        head_end      : str
        text          : str
        head_case     : str                                   = cls.section(0).case
        body_case     : str                                   = cls.section(1).case
        head_text     : set[str]                              = set()
        body_text     : list[str]                             = []
        inst_data     : InstanceData                          = {}
        post_data     : PostInstanceData                      = {}
        marks         : type[StrangAPI.StrangMarkAbstract_e]  = cls.Marks  # type: ignore[attr-defined]
        ctor          : Ctor[T]                               = cls
        match cls.section(0).end:
            case None:
                raise ValueError()
            case str() as x:
                head_end                                      = x

        match marks.default():
            case None:
                raise ValueError()
            case x:
                default_mark  = marks.default()
        ##--| clean the input
        match input:
            case StrangAPI.Strang_p() as val:
                x, m, y = val[:,:].partition(head_end)
                if bool(m):
                    head_text.update(x.split(head_case))
                    body_text.append(y)
                else:
                    body_text.append(x)
            case pl.Path() as val if bool(val.suffix):
                head_text.add(marks.file)
                body_text.append(str(val))
            case str() as val:
                x, m, y = val.partition(head_end)
                if bool(m):
                    head_text.update(x.split(head_case))
                    body_text.append(y)
                else:
                    body_text.append(x)
            case val:
                x, m, y = str(val).partition(head_end)
                if bool(m):
                    head_text.update(x.split(head_case))
                    body_text.append(y)
                else:
                    body_text.append(x)


        body : str = body_case.join(body_text)
        if any(x.value in body for x in cls.section(1).marks):
            head_text.add(marks.abstract.value)
        if "." in body:
            head_text.add(marks.file.value)

        if not bool(head_text):
            head_text.add(default_mark)
        assert(bool(body_text))
        text = head_end.join([head_case.join(head_text), body])
        return text, inst_data, post_data, ctor




#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from time import sleep
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv.structs.chainguard.errors import GuardedAccessError

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv.structs.chainguard._base import TomlTypes, GuardBase
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

super_get             = object.__getattribute__
super_set             = object.__setattr__

class TomlAccess_m:
    """ Mixing for dynamic attribute access """

    def __setattr__(self, attr:str, value:TomlTypes) -> None:
        if not getattr(self, "__mutable"):
            raise TypeError()
        super_set(self, attr, value)

    def __getattr__(self, attr:str) -> GuardBase | TomlTypes | list[GuardBase]:
        table = self._table()

        if attr not in table and attr.replace("_", "-") not in table:
            index     = self._index() + [attr]
            index_s   = ".".join(index)
            available = " ".join(self.keys())
            raise GuardedAccessError(f"{index_s} not found, available: [{available}]")

        match table.get(attr, None) or table.get(attr.replace("_", "-"), None):
            case dict() as result:
                return self.__class__(result, index=self._index() + [attr])
            case list() as result if all(isinstance(x, dict) for x in result):
                index = self._index()
                return [self.__class__(x, index=index[:]) for x in result if isinstance(x, dict)]
            case _ as result:
                return result

    def __getitem__(self, keys:str|list[str]|tuple[str]) -> TomlTypes:
        curr : Self = self
        match keys:
            case tuple():
                for key in keys:
                    curr = curr.__getattr__(key)
            case str():
                curr = self.__getattr__(keys)
            case _:
                pass

        return curr

    def get(self, key:str, default:Maybe[TomlTypes]=None) -> Maybe[TomlTypes]:
        if key in self:
            return self[key]

        return default



#!/usr/bin/env python3
"""



"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

from jgdv.errors import JGDVError

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Global Vars:

# Body:
class ParseError(JGDVError):
    """ A Base Error Class for JGDV CLI Arg Parsing"""
    pass

class HeadParseError(ParseError):
    """ For When an error occurs parsing the head """
    pass

class CmdParseError(ParseError):
    """ For when parsing the command section fails """
    pass

class SubCmdParseError(ParseError):
    """ For when the subcmd section fails """
    pass

class ArgParseError(ParseError):
    """ For when a head/cmd/subcmds arguments are bad """
    pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: N801, ANN001, ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from jgdv import identity_fn
from jgdv.structs.strang import _interface as StrangAPI # noqa: N812

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Mapping

if TYPE_CHECKING:
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

    from jgdv import Maybe, Rx, Ident, RxStr, Ctor, CHECKTYPE, FmtStr
    from ._util._interface import Expander_p, ExpInst_d

    type LitFalse    = Literal[False]
    type KeyMark     = DKeyMarkAbstract_e|LitFalse|str|type|tuple[KeyMark, ...]
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
DEFAULT_COUNT            : Final[int]              = 0
FMT_PATTERN              : Final[Rx]               = re.compile(r"[wdi]+")
EXPANSION_LIMIT_PATTERN  : Final[Rx]               = re.compile(r"e(\d+)")
INDIRECT_SUFFIX          : Final[Ident]            = "_"
KEY_PATTERN              : Final[RxStr]            = "{(.+?)}"
OBRACE                   : Final[str]              = "{"
MAX_DEPTH                : Final[int]              = 10
MAX_KEY_EXPANSIONS       : Final[int]              = 200
PAUSE_COUNT              : Final[int]              = 0
RECURSION_GUARD          : Final[int]              = 10
PARAM_IGNORES            : Final[tuple[str, str]]  = ("_", "_ex")

RAWKEY_ID                : Final[str]              = "_rawkeys"
FORCE_ID                 : Final[str]              = "force"
ARGS_K                   : Final[Ident]            = "args"
KWARGS_K                 : Final[Ident]            = "kwargs"

DEFAULT_DKEY_KWARGS      : Final[list[str]]        = [
    "ctor", "check", "mark", "fallback",
    "max_exp", "fmt", "help", "implicit", "conv",
    "named",
    RAWKEY_ID, FORCE_ID,
    ]

##--| Error Messages
UnknownDKeyCtorType     : Final[str]  = "Unknown type passed to construct dkey"
InsistentKeyFailure     : Final[str]  = "An insistent key was not built"
KeyBuildFailure         : Final[str]  = "No key was built"
NoMark                  : Final[str]  = "Mark has to be a value"
MarkConflictsWithMulti  : Final[str]  = "Mark is MULTI but multi=False"
MarkLacksACtor          : Final[str]  = "Couldn't find a ctor for mark"
MarkConversionConflict  : Final[str]  = "Kwd Mark/Conversion Conflict"
UnexpectedKwargs        : Final[str]  = "Key got unexpected kwargs"
RegistryLacksMark       : Final[str]  = "Can't register when the mark is None"
RegistryConflict        : Final[str]  = "API.Key_p Registry conflict"
ConvParamTooLong        : Final[str]  = "Conversion Parameters For Dkey's Can't Be More Than A Single Char"
ConvParamConflict       : Final[str]  = "Conversion Param Conflict"
# Enums:

class DKeyMarkAbstract_e(StrangAPI.StrangMarkAbstract_e):

    @classmethod
    def default(cls) -> Maybe: ...

    @classmethod
    def null(cls) -> Maybe: ...

    @classmethod
    def multi(cls) -> Maybe: ...

class DKeyMark_e(DKeyMarkAbstract_e):
    """
      Enums for how to use/build a dkey

    """
    ARGS     = enum.auto() # -> list
    KWARGS   = enum.auto() # -> dict
    POSTBOX  = enum.auto() # -> list

    @classmethod
    def default(cls) -> Any:  # noqa: ANN401
        return Any

    @classmethod
    def null(cls) -> Maybe:
        return False

    @classmethod
    def indirect(cls) -> type:
        return Mapping

    @classmethod
    def multi(cls) -> type:
        return list

##--| Data

class RawKey_d:
    """ Utility class for parsed {}-format string parameters.

    ::

        see: https://peps.python.org/pep-3101/
        and: https://docs.python.org/3/library/string.html#format-string-syntax

    Provides the data from string.Formatter.parse, but in a structure
    instead of a tuple.
    """
    __slots__ = ("convert", "format", "key", "prefix")
    prefix  : str
    key     : Maybe[str]
    format  : Maybe[str]
    convert : Maybe[str]

    def __init__(self, **kwargs) -> None:
        self.prefix       = kwargs.pop("prefix")
        self.key          = kwargs.pop("key", None)
        self.format       = kwargs.pop("format", None)
        self.convert      = kwargs.pop("convert", None)
        assert(not bool(kwargs)), kwargs

    def __getitem__(self, i) -> Maybe[str]:
        match i:
            case 0:
                return self.prefix
            case 1:
                return self.key
            case 2:
                return self.format
            case 3:
                return self.convert
            case _:
                msg = "Tried to access a bad element of DKeyParams"
                raise ValueError(msg, i)

    def __bool__(self) -> bool:
        return bool(self.key)

    def __repr__(self) -> str:
        return f"<RawkKey: {self.joined()}>"

    def joined(self) -> str:
        """ Returns the key and params as one string

        eg: blah, fmt=5, conv=p -> blah:5!p
        """
        args : list[str]
        if not bool(self.key):
            return ""

        assert(self.key is not None)
        args = [self.key]
        if bool(self.format):
            assert(self.format is not None)
            args += [":", self.format]
        if bool(self.convert):
            assert(self.convert is not None)
            args += ["!", self.convert]

        return "".join(args)

    def wrapped(self) -> str:
        """ Returns this key in simple wrapped form

        (it ignores format, conv params and prefix)

        eg: blah -> {blah}
        """
        return "{%s}" % self.key  # noqa: UP031

    def anon(self) -> str:
        """ Make a format str of this key, with anon variables.

        eg: blah {key:f!p} -> blah {}
        """
        if bool(self.key):
            return "%s{}" % self.prefix  # noqa: UP031

        return self.prefix or ""

    def direct(self) -> str:
        """ Returns this key in direct form

        ::

            eg: blah -> blah
                blah_ -> blah
        """
        return (self.key or "").removesuffix(INDIRECT_SUFFIX)

    def indirect(self) -> str:
        """ Returns this key in indirect form

        ::

            eg: blah -> blah_
                blah_ -> blah_
        """
        match self.key:
            case str() as k if k.endswith(INDIRECT_SUFFIX):
                return k
            case str() as k:
                return f"{k}{INDIRECT_SUFFIX}"
            case _:
                return ""

    def is_indirect(self) -> bool:
        match self.key:
            case str() as k if k.endswith(INDIRECT_SUFFIX):
                return True
            case _:
                return False

class DKey_d(StrangAPI.Strang_d):
    """ Data of a DKey """
    __slots__ = ("convert", "expansion_type", "fallback", "format", "help", "max_expansions", "multi", "name", "raw", "typecheck")
    name            : Maybe[str]
    raw             : tuple[RawKey_d, ...]
    expansion_type  : Ctor
    typecheck       : CHECKTYPE
    fallback        : Maybe[Any]
    format          : Maybe[FmtStr]
    convert         : Maybe[FmtStr]
    help            : Maybe[str]
    max_expansions  : Maybe[int]
    multi           : bool

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.name            = kwargs.pop("name", None)
        self.raw             = tuple(kwargs.pop(RAWKEY_ID, ()))
        self.expansion_type  = kwargs.pop("ctor", identity_fn)
        self.typecheck       = kwargs.pop("check", Any)
        self.fallback        = kwargs.pop("fallback", None)
        self.format          = kwargs.pop("format", None)
        self.convert         = kwargs.pop("convert", None)
        self.help            = kwargs.pop("help", None)
        self.max_expansions  = kwargs.pop("max_exp", None)
        self.multi           = kwargs.pop("multi", False)

##--| Section Specs
DKEY_SECTIONS : Final[StrangAPI.Sections_d] = StrangAPI.Sections_d(
    StrangAPI.Sec_d("body", None, None, str, None, True),  # noqa: FBT003
)
##--| Protocols

@runtime_checkable
class Key_p(StrangAPI.Strang_p, Protocol):
    """ The protocol for a Key, something that used in a template system"""
    _extra_kwargs  : ClassVar[set[str]]
    _processor     : ClassVar
    _expander      : ClassVar[Expander_p]
    data           : DKey_d

    @staticmethod
    def MarkOf[T:Key_p](cls:type[T]|T) -> KeyMark: ...  # noqa: N802, PLW0211

    def redirect(self, spec=None) -> Key_p: ...

    def expand(self, *sources, rec=False, insist=False, chain:Maybe[list[Key_p]]=None, on_fail=Any, locs:Maybe[Mapping]=None, **kwargs) -> str: ...

    def var_name(self) -> str: ...

    def to_exp_inst(self, *, indirect:bool=False, **kwargs:Any) -> ExpInst_d: ...  # noqa: ANN401

@runtime_checkable
class MultiKey_p(Protocol):

    def keys(self) -> list[Key_p]: ...

    def _multi(self) -> Literal[True]: ...

@runtime_checkable
class IndirectKey_p(Protocol):

    def _indirect(self) -> Literal[True]: ...
##--| Combined Interfaces

@runtime_checkable
class NonKey_p(Protocol):

    def _nonkey(self) -> Literal[True]: ...



"""
This module provides some miscellaneous mixin classes.

SubAnnotate_m and SubRegistry_m allow for creating annotated subclasses.
(eg: Strang[int])

Human_m is for rounding time and formatting large numbers easier.

PathManip_m is a simple toolbox for some path operations I keep using.

Zipper_m is for zipfile operations.

"""




#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast, Self, Final)
import warnings

import pytest

logging = logmod.root

from ... import DKey, Key_p
from ..str_key import StrDKey


IMP_KEY_BASES               : Final[list[str]]           = ["bob", "bill", "blah", "other", "23boo", "aweg2531", "awe_weg", "aweg-weji-joi"]
EXP_KEY_BASES               : Final[list[str]]           = [f"{{{x}}}" for x in IMP_KEY_BASES]
EXP_P_KEY_BASES             : Final[list[str]]           = ["{bob:wd}", "{bill:w}", "{blah:wi}", "{other:i}"]
PATH_KEYS                   : Final[list[str]]           = ["{bob}/{bill}", "{blah}/{bloo}", "{blah}/{bloo}"]
MUTI_KEYS                   : Final[list[str]]           = ["{bob}_{bill}", "{blah} <> {bloo}", "! {blah}! {bloo}!"]
IMP_IND_KEYS                : Final[list[str]]           = ["bob_", "bill_", "blah_", "other_"]
EXP_IND_KEYS                : Final[list[str]]           = [f"{{{x}}}" for x in IMP_IND_KEYS]

VALID_KEYS                                           = IMP_KEY_BASES + EXP_KEY_BASES + EXP_P_KEY_BASES + IMP_IND_KEYS + EXP_IND_KEYS
VALID_MULTI_KEYS                                     = PATH_KEYS + MUTI_KEYS

class TestStrKey:

    def test_mark(self):
        assert(DKey.MarkOf(StrDKey) is str)

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# mypy: disable-error-code="attr-defined"
# ruff: noqa: ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from pydantic import BaseModel, field_validator, model_validator

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Proto
from jgdv.structs.dkey import DKey
from jgdv.mixins.path_manip import PathManip_m
from jgdv.structs.strang import Strang

from jgdv.structs.strang import _interface as StrangAPI # noqa: N812
from jgdv.structs.strang._interface import Strang_p
# ##-- end 1st party imports

from . import _interface as API # noqa: N812
from .processor import LocationProcessor

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
TimeDelta = datetime.timedelta
if TYPE_CHECKING:
   import enum
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.Location_p)
class Location(Strang):
    """ A Location is an abstraction higher than a path.

    ie: a path, with metadata.

    Doesn't expand on its own, requires a JGDVLocator store

    It is a Strang subclass, of the form "{meta}+::a/path/location". eg::

        file/clean::.temp/docs/blah.rst

    TODO use annotations to require certain metaflags.
    eg::

        ProtectedLoc = Location['protect']
        Cleanable    = Location['clean']
        FileLoc      = Location['file']

    TODO add an ExpandedLoc subclass that holds the expanded path,
    and removes the need for much of PathManip_m?

    TODO add a ShadowLoc subclass using annotations
    eg::

        BackupTo           = ShadowLoc[root='/vols/BackupSD']
        a_loc              = BackupTo('file::a/b/c.mp3')
        a_loc.path_pair() -> ('/vols/BackupSD/a/b/c.mp3', '~/a/b/c.mp3')

    """
    __slots__              = ()

    _processor : ClassVar  = LocationProcessor()
    _sections  : ClassVar  = API.LocationSections
    Marks      : ClassVar  = API.LocationMeta_e
    Wild       : ClassVar  = API.WildCard_e

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs) # type: ignore[misc]

    def __contains__(self, other:object) -> bool: # type: ignore
        """ Whether a definite artifact is matched by self, an abstract artifact

       | other    ∈ self
       | a/b/c.py ∈ a/b/*.py
       | ________ ∈ a/*/c.py
       | ________ ∈ a/b/c.*
       | ________ ∈ a/*/c.*
       | ________ ∈ **/c.py
       | ________ ∈ a/b ie: self < other
        """
        match other:
            case self.Marks() as x:
                return x in self.data.meta
            case pl.Path() | API.Location_p() if self.Marks.abstract in self.data.meta:
                return self.check_wildcards(other)
            case API.Location_p():
                return self < other
            case _:
                return super().__contains__(other) # type: ignore[misc]

    @override
    def __lt__(self, other:TimeDelta|str|pl.Path|Location) -> bool: # type: ignore[override]
        """ self < path|location
            self < delta : self.modtime < (now - delta)
        """
        match other:
            case TimeDelta() if self.is_concrete():
                return False
            case TimeDelta():
                raise NotImplementedError()
            case _:
                return super().__lt__(str(other)) # type: ignore[misc]

    @property
    def path(self) -> pl.Path: # type: ignore
        return pl.Path(self[1,:])

    @property
    def body_parent(self) -> list[str|API.WildCard_e]:
        if self.Marks.file in self:
            return list(itz.islice(self.words(1), len(self.data.sec_words[1])-1))

        return list(self.words(1))

    @property
    def stem(self) -> Maybe[str|tuple[API.WildCard_e, str]]: # type: ignore
        """ Return the stem, or a tuple describing how it is a wildcard """
        result : Maybe[str|tuple[API.WildCard_e, str]] = None
        if self.Marks.file not in self.data.meta:
            return result

        match self[1,-1]:
            case str() as elem:
                pass
            case _:
                return None

        match elem.split(".")[0]:
            case str() as elem if (wc:=self.Wild.glob) in elem:
                result = (wc, elem)
            case str() as elem if (wc:=self.Wild.select) in elem:
                result = (wc, elem)
            case str() as elem if (wc:=self.Wild.key) in elem:
                result = (wc, elem)
            case str() as elem:
                result = elem
            case _:
                pass

        return result

    @property
    def keys(self) -> set[str]:
        raise NotImplementedError()

    @property
    def key(self) -> Maybe[str|API.Key_p]:
        raise NotImplementedError()

    def ext(self, *, last:bool=False) -> Maybe[str|tuple[API.WildCard_e, str]]: # type: ignore # noqa: PLR0911
        """ return the ext, or a tuple of how it is a wildcard.
        returns nothing if theres no extension,
        returns all suffixes if there are multiple, or just the last if last=True
        """
        if self.Marks.file not in self.data.meta:
            return None

        match self[1,-1]:
            case str() as elem:
                pass
            case _:
                return None

        match elem.rfind(".") if last else elem.find("."):
            case -1:
                return None
            case int() as x:
                pass

        match elem[x:]:
            case ".":
                return None
            case ext if (wc:=API.WildCard_e.glob) in ext:
                return (wc, ext)
            case ext if (wc:=API.WildCard_e.select) in ext:
                return (wc, ext)
            case ext:
                return ext

    def is_concrete(self) -> bool:
        return self.Marks.abstract not in self.data.meta

    def check_wildcards(self, other:pl.Path|API.Location_p) -> bool:  # noqa: PLR0912
        """ Return True if other is within self, accounting for wildcards """
        logging.debug("Checking %s < %s", self, other)
        x            : Any
        other_ext    : Any
        other_parts  : list[str]
        wildcards    : set[str]
        if self.is_concrete():
            return self < other

        match other:
            case pl.Path() as x:
                other_parts  = list(x.parts)
                other_ext    = x.suffix
            case API.Location_p() as x:
                other_parts = other.body_parent
                other_ext = other.ext()

        wildcards = {x.value for x in self.section(1).marks}
        # Compare path up to the file
        for x,y in zip(self.body_parent, other_parts, strict=False):
            match x, y:
                case str(), str() if x == y:
                    pass
                case str(), str() if x not in self.Wild and y not in self.Wild:
                    return False
                case x, _ if self.Wild(x) is self.Wild.rec_glob:
                    break
                case x, str() if x in self.Wild:
                    pass
                case str(), y if y in self.Wild:
                    pass
                case str(), str():
                    pass

        if self.Marks.file not in self.data.meta:
            return True

        logging.debug("%s and %s match on path", self, other)
        # Compare the stem/ext
        match self.stem, other.stem:
            case (xa, ya), (xb, yb) if xa == xb and ya == yb:
                pass
            case (xa, ya), str():
                pass
            case str() as x, str() as y if x == y:
                pass
            case _, _:
                return False

        logging.debug("%s and %s match on stem", self, other)
        match self.ext(), other_ext:
            case None, None:
                pass
            case (xa, ya), (xb, yb) if xa == xb and ya == yb:
                pass
            case (x, y), _:
                pass
            case str() as x, str() as y if x == y:
                pass
            case _, _:
                return False

        logging.debug("%s and %s match", self, other)
        return True



"""
:class:`JGDVLogConfig` provides a single config point for logging,
both to files and stdout/error. The Log Config is a singleton.

Different loggers can be configured using TOML,
which is parsed into the structure :class:`LoggerSpec`.

The module also provides some logging extensions:
- :class:`JGDVLogger`
- :class:`JGDVLogRecord`
- :class`LogCall`, a decorator.

"""

from ._interface import LogLevel_e
from .config import JGDVLogConfig



#!/usr/bin/env python3
"""
TEST File updated

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from ..util_decorators import DoMaybe


# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# from . import _interface as API # noqa: N812
# Dataclasses:
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError


#
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:
class TestSuite:

    ##--|
    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133


    def test_basic(self) -> None:

        class Example:

            @DoMaybe()
            def bfunc(self, val:int) -> int:
                return val

        from typing import reveal_type
        reveal_type(Example.bfunc)

        assert(True)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
import timeit
# ##-- end stdlib imports

import jgdv

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Traceback
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger    = logmod.Logger
    type Statement = str | Callable
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

# @Proto(CtxManager_p)
class MultiTimeBlock_ctx:
    """ CtxManager for timing statements multiple times

    see https://docs.python.org/3/library/timeit.html
    """
    count              : int
    repeat             : int
    keep_gc            : bool
    group              : str
    autorange_total    : float
    once_log           : list[tuple[str, float]]
    log_level          : int
    _logger            : Maybe[Logger]
    _current_name      : Maybe[str]

    def __init__(self, *, count:int=10, repeat:int=5, keep_gc:bool=False, group:Maybe[str]=None, logger:Maybe[Logger]=None, level:Maybe[int|str]=None) -> None:  # noqa: PLR0913
        self.count            = count
        self.repeat           = repeat
        self.keep_gc          = keep_gc
        self.group : str      = group if group else ""
        self.autorange_total  = 0.0
        self.once_log         = []
        self._logger          = logger
        match level:
            case None:
                self.log_level = logmod._nameToLevel["info"]
            case int() as x:
                self.log_level = x
            case str():
                self.log_level            = logmod._nameToLevel[level or "info"]

    def _set_name(self, *, name:str, stmt:Callable) -> None:
        """ Default Name builder """
        match name, stmt:
            case str(), str():
                self._current_name = f"{self.group}::{name}"
            case str(), x if hasattr(x, "__qualname__"):
                self._current_name = "::".join(self.group, name, stmt.__qualname__) #type:ignore
            case str(), _:
                self._current_name = f"{self.group}::{name}"

    def autorange_cb(self, number:int, took:float) -> None:
        """ Callback for autorange.
        Called after each trial.
        """
        self._log("%-*10s : %-*5d calls took:", self._current_name, number, time=took)
        self.autorange_total += took

    def auto(self, stmt:Callable, *, name:Maybe[str]=None) -> float:
        """ Try the statement with larger trial sizes until it takes at least 0.2 seconds """
        self._set_name(name=name, stmt=stmt)
        self._log("Autoranging: %s", self._current_name)
        timer = timeit.Timer(stmt, globals=globals())
        timer.autorange(self.autorange_cb)
        return self.autorange_total

    def repeats(self, stmt:Callable, *, name:Maybe[str]=None) -> list[float]:
        """
        Repeat the stmt and report the results
        """
        self._set_name(name=name, stmt=stmt)
        self._log("Repeating %s : Timing %s repeats of %s trials",
                  self._current_name, self.repeat, self.count)
        timer  = timeit.Timer(stmt, globals=globals())
        results = timer.repeat(repeat=self.repeat, number=self.count)
        for i, result in enumerate(results):
            self._log("Attempt %-*5d : %-*8.2f seconds", i, time=result, prefix="----")
        else:
            return results

    def block(self, stmt:Callable, *, name:Maybe[str]=None) -> float:
        """ Run the stmt {count} numnber of times and report the time it took"""
        self._set_name(name=name, stmt=stmt)
        self._log("Running Block %s : Timing block of %-*5f trials",
                  self._current_name, self.count)
        timer  = timeit.Timer(stmt, globals=globals())
        result = timer.timeit(self.count)
        self._log("%-*10s : %-*8.2f seconds", self._current_name, time=result, prefix="----")
        return result

    def once(self, stmt:Callable, *, name:Maybe[str]=None) -> float :
        """ Run the statement once, and return the time it took """
        self._set_name(name=name, stmt=stmt)
        self._log("Running Call Once: %s", self._current_name)
        timer  = timeit.Timer(stmt, globals=globals())
        result = timer.timeit(1)
        self.once_log.append((self._current_name, result))
        self._log("%-*10s", self._current_name, time=result, prefix="----")
        return result

    def _log(self, msg:str, *args:Any, time:Maybe[float]=None, prefix:Maybe[str]=None) -> None:  # noqa: ANN401
        """ The internal log method """
        match self._logger:
            case None:
                pass
            case logmod.Logger() if time is None:
                prefix = prefix or "--"
                msg_format = f"%s {msg}"
                self._logger.log(self.log_level, msg_format, prefix, *args)
            case logmod.Logger():
                prefix = prefix or "--"
                msg_format = f"%s {msg} : %-*8.2f seconds"
                self._logger.log(self.log_level, msg_format, prefix, *args, time)

    def __enter__(self) -> Self:
        """ Return a copy of this obj for a with block """
        match self.group:
            case str() as x:
                self._log("Entering: %s", x)
            case None:
                pass
        return deepcopy(self)

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        """ On exiting the block, reports the time the block took """
        match self.autorange_total:
            case float()as x if 0 < x:
                self._log("Finished Block %s", self.group, time=x)
            case _:
                self._log("Finished Block %s", self.group)

        match self.once_log:
            case []:
                pass
            case [*xs]:
                long_name, time_taken = max(xs, key=lambda x: x[1])
                self._log("-- Longest Single Call: %s : %-*8.2fs seconds", long_name, time_taken )

        return False



"""


"""
from __future__ import annotations

# Import:
from typing import Final
from jgdv.debugging import TraceBuilder
from jgdv.errors import JGDVError

##--| Error Messages
MissingSectionName              : Final[str]  = "{cls} has no section {key}"
MissingSectionIndex             : Final[str]  = "{cls} has no section {idx}, only {sections}"
SliceMisMatch                   : Final[str]  = "Mismatch between section slices and word slices"
UnkownSlice                     : Final[str]  = "Slice Logic Failed"
IndexOfEmptyStr                 : Final[str]  = "Tried to Index the empty str"
MalformedData                   : Final[str]  = "Base data malformed"
TooManyUUIDs                    : Final[str]  = "More than One UUID found"
NoUUIDToDiff                    : Final[str]  = "Tried to diff a uuid with a non-uniq strang"
StrangCtorFailure               : Final[str]  = "[{cls}] Stage: {stage}"
CodeRefImportFailed             : Final[str]  = "Attempted import failed, attribute not found"
CodeRefImportNotCallable        : Final[str]  = "Imported 'Function' was not a callable"
CodeRefImportNotClass           : Final[str]  = "Imported 'Class' was not a type"
CodeRefImportCheckFail          : Final[str]  = "Imported Value does not match required type"
CodeRefImportUnknownFail        : Final[str]  = "Imported Code Reference is not of correct type"
CodeRefImportNotValue           : Final[str]  = "Imported Code Reference is not a value"
CodeRefUUIDFail                 : Final[str]  = "Code References shouldn't need UUIDs"

FormatterExpansionTypeFail      : Final[str]  = "Unrecognized expansion type"
FormatterConversionUnknownSpec  : Final[str]  = "Unknown conversion specifier {spec!s}"
FormatterUnkownBodyType         : Final[str]  = "Unknown body type"

NoSkipMark                      : Final[str]  = "Can't push without a skip mark in the last section"


##--| Error Classes
class StrangError(JGDVError):
    """ The Base Error type that Strang's raise """
    pass



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
from types import MethodType
import warnings

import pytest

from jgdv import JGDVError
from jgdv.decorators import DForm_e
from ... import DKey, DKeyed
from ..decorator import DKeyMetaDecorator, DKeyedMeta, DKeyedRetrieval
from ..decorator import DKeyExpansionDecorator as DKexd
from ...errors import DecorationMismatch

logging  = logmod.root

KEY_1    : Final[str] =  "blah"
KEY_2    : Final[str] =  "bloo"
KEY_W1   : Final[str] =  "{blah}"
KEY_W2   : Final[str] =  "{bloo}"

class TestDkeyDecorator:

    def test_sanity(self):
        assert(True is not False)

    def test_initial(self):
        value = DKexd([])
        assert(isinstance(value, DKexd))

    def test_with_key(self):
        value = DKexd([DKey(KEY_1)])
        assert(isinstance(value, DKexd))
        assert(bool(value._data))

    def test_validate_sig_method_head(self):
        dec   = DKexd([])
        ttype = DForm_e.METHOD

        def simple(self, spec, state):
            pass

        sig = dec._signature(simple)
        dec._validate_sig_h(sig, ttype, [])
        assert(True)

    def test_validate_sig_method_head_fail(self):
        dec   = DKexd([])
        ttype = DForm_e.METHOD

        def simple(self, spec, notstate):
            pass

        sig = dec._signature(simple)
        with pytest.raises(JGDVError):
            dec._validate_sig_h(sig, ttype, [])

    def test_validate_sig_func_head(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, state):
            pass

        sig = dec._signature(simple)
        dec._validate_sig_h(sig, ttype, [])
        assert(True)

    def test_validate_sig_func_head_fail(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, notstate):
            pass

        sig = dec._signature(simple)
        with pytest.raises(JGDVError):
            dec._validate_sig_h(sig, ttype, [])

    def test_validate_sig_method_tail(self):
        dec   = DKexd([])
        ttype = DForm_e.METHOD

        def simple(self, spec, state, bloo, blee):
            pass

        sig = dec._signature(simple)
        with pytest.raises(DecorationMismatch):
            dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2)])

    def test_validate_sig_method_tail_fail(self):
        dec   = DKexd([])
        ttype = DForm_e.METHOD

        def simple(self, spec, state, bloo, blee):
            pass

        sig = dec._signature(simple)
        with pytest.raises(JGDVError):
            dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2)])

    def test_validate_sig_func_tail(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, state, blah, bloo):
            pass

        sig = dec._signature(simple)
        dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2)])
        assert(True)

    def test_validate_sig_func_tail_fail(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, state, bloo, blee):
            pass

        sig = dec._signature(simple)
        with pytest.raises(JGDVError):
            dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2)])

    def test_validate_sig_incomplete_tail(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, state, bloo, blee, blob):
            pass

        sig = dec._signature(simple)
        with pytest.raises(DecorationMismatch):
            dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2)])

    def test_validate_sig_skip_ignores(self):
        dec   = DKexd([])
        ttype = DForm_e.FUNC

        def simple(spec, state, _bloo, blee_ex, blob):
            pass

        sig = dec._signature(simple)
        dec._validate_sig_h(sig, ttype, [DKey(KEY_W1), DKey(KEY_W2), DKey("{blob}")])
        assert(True)

class TestDKeyDecoratorExpansion:

    def test_sanity(self):
        assert(True is not False)

    def test_basic(self):
        state = { "basic": "blah" }

        @DKeyed.types("basic")
        def simple(spec, state, basic):
            assert(basic == "blah")

        simple(None, state)

    def test_mismatch_signature(self):
        with pytest.raises(JGDVError):
            @DKeyed.types("other")
            def simple(spec, state, basic):
                assert(basic == "blah")

    def test_expansion_fallback(self):
        state = { "notbasic": "bloo" }

        @DKeyed.types("basic", fallback="blah")
        def simple(spec, state, basic):
            assert(basic == "blah")

        simple(None, state)

    def test_multi_expansion(self):
        state = { "basic": "bloo", "other": "qwerty" }

        @DKeyed.types("basic", fallback="blah")
        @DKeyed.types("other", fallback="aweg")
        def simple(spec, state, basic, other):
            assert(basic == "bloo")
            assert(other == "qwerty")

        simple(None, state)

    def test_multi_expansion_fallback(self):
        state = { "notbasic": "bloo", "notother": "qwerty" }

        @DKeyed.types("basic", fallback="blah")
        @DKeyed.types("other", fallback="aweg")
        def simple(spec, state, basic, other):
            assert(basic == "blah")
            assert(other == "aweg")

        simple(None, state)

    def test_meta_decorator_no_change(self):

        @DKeyed.requires("basic")
        @DKeyed.requires("other")
        def simple(spec, state):
            pass

        simple(None, None)


    def test_redirects_expansion(self):

        @DKeyed.redirects("other_")
        def simple(spec, state, _other):
            match _other:
                case str():
                    assert(_other == "blah")
                case x:
                    assert(False), x

        simple(None, {"other_":"blah"})


    def test_redirects_expansion_doesnt_recurse(self):
        """
        other_ -> {blah} -X-> bloo
        """

        @DKeyed.redirects("other_")
        def simple(spec, state, _other):
            match _other:
                case DKey():
                    assert(_other == "blah")
                case x:
                    assert(False), x

        simple(None, {"other_":"blah", "blah":"bloo"})

class TestDKeyed:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        assert(DKeyedMeta in DKeyed._extensions)
        assert(DKeyedRetrieval in DKeyed._extensions)

    def test_simple_access(self):
        assert(hasattr(DKeyedMeta, "requires"))
        assert(hasattr(DKeyed, "requires"))
        match DKeyed.requires:
            case MethodType():
                assert(True)
            case x:
                assert(False), x


    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import tracemalloc
import types
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import DateTime, Seconds
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class HumanNumbers_m:
    """
    Simple Mixin for human related functions
    """

    @staticmethod
    def humanize(val:int|float, *, force_sign:bool=True) -> str:
        """ Format {val} in a human readable way as a size.
        Uses tracemalloc._format_size.
        Depending on size, will use on of the units:
        B, KiB, MiB, GiB, TiB.
        """
        return tracemalloc._format_size(val, force_sign) # type:ignore[attr-defined]


    @staticmethod
    def round_time(dt:DateTime=None, *,  round:Seconds=60) -> DateTime:
        """Round a datetime object to any time lapse in seconds
        dt : datetime.datetime object, default now.
        round : Closest number of seconds to round to, default 1 minute.
        Author: Thierry Husson 2012 - Use it as you want but don't blame me.
        from: https://stackoverflow.com/questions/3463930
        """
        dt       = dt or datetime.datetime.now()
        seconds  = (dt.replace(tzinfo=None) - dt.min).seconds
        rounding = (seconds+round/2) // round * round
        return dt + datetime.timedelta(0,rounding-seconds,-dt.microsecond)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011, PLR2004

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from jgdv.structs.strang import Strang
from jgdv.structs.strang.errors import StrangError
from ..dkey import DKey
from ..keys import SingleDKey, MultiDKey, NonDKey, IndirectDKey

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from typing import Any
from collections.abc import Mapping

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestSingleDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        assert(DKey.MarkOf(SingleDKey) == Any)

    def test_basic(self):
        match DKey("blah", implicit=True, force=SingleDKey):
            case SingleDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, SingleDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_basic_by_class_accesss(self):
        match DKey[Any]("blah", implicit=True):
            case SingleDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, SingleDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_fail_on_multiple_keys(self):
        with pytest.raises(StrangError):
            DKey("{blah} awef {bloo}", force=SingleDKey)

    def test_eq(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = DKey("blah", implicit=True, force=SingleDKey)
        assert(obj1 == obj2)

    def test_eq_str(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = "blah"
        assert(obj1 == obj2)

    def test_eq_not_implemented(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = 21
        assert(not (obj1 == obj2))

    def test_hash(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = "blah"
        assert(hash(obj1) == hash(obj2))

    def test_str(self):
        obj1 = DKey("{blah}")
        obj2 = "blah"
        assert(obj1 == obj2)

    def test_format_wrapped(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = "{blah}"
        assert(f"{obj1:w}" == obj2)

    def test_format_indirect(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = "blah_"
        assert(f"{obj1:i}" == obj2)

    def test_format_indirect_wrapped(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        obj2 = "{blah_}"
        assert(f"{obj1:wi}" == obj2)

    def test_getitem_key_fails(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        with pytest.raises(IndexError):
            obj1[0,0]

    def test_get_key_basic(self):
        obj1 = DKey("blah", implicit=True, force=SingleDKey)
        with pytest.raises(IndexError):
            obj1.get(0,0)

    def test_get_key_from_wrapped(self, force=SingleDKey):
        obj1 = DKey("{blah}")
        with pytest.raises(IndexError):
            obj1.get(0,0)

class TestMultiDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        match DKey("{blah} {bloo}", force=MultiDKey):
            case MultiDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, MultiDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_basic_by_class_access(self):
        match DKey[list]("{blah} {bloo}"):
            case MultiDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, MultiDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_basic_unannotated(self):
        match DKey("{blah} {bloo}"):
            case MultiDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, MultiDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_subkey_expansion_control(self):
        obj = DKey("{blah:e1} {bloo:e2}")
        match obj.keys():
            case [x, y]:
                assert(x.data.max_expansions == 1)
                assert(y.data.max_expansions == 2)
            case x:
                assert(False), x

    def test_eq(self):
        obj1 = DKey("{blah} {bloo}", force=MultiDKey)
        obj2 = DKey("{blah} {bloo}", force=MultiDKey)
        assert(obj1 == obj2)

    def test_eq_str(self):
        obj1 = DKey("{blah} {bloo}", force=MultiDKey)
        obj2 = "{blah} {bloo}"
        assert(obj1 == obj2)

    def test_eq_not_implemented(self):
        obj1 = DKey("{blah} {bloo}", force=MultiDKey)
        obj2 = 21
        assert(not (obj1 == obj2))

    def test_subkeys(self):
        obj = DKey("{first} {second} {third}", force=MultiDKey)
        assert(isinstance(obj, MultiDKey))
        for sub in obj.keys():
            assert(isinstance(sub, SingleDKey)), type(sub)

    def test_subkeys_two(self):
        obj = DKey("{test} then {blah}")
        assert(isinstance(obj, MultiDKey))
        assert(len(obj.keys()) == 2)

    def test_anon(self):
        obj = DKey("{first} {second} {third}", force=MultiDKey)
        assert(isinstance(obj, MultiDKey))
        assert(obj.anon == "{} {} {}")

    def test_anon_2(self):
        obj = DKey[list]("{b}")
        assert(isinstance(obj, MultiDKey))
        assert(obj.anon == "{}")

    def test_hash(self):
        obj1 = DKey("{blah}", force=MultiDKey)
        assert(isinstance(obj1, MultiDKey))
        obj2 = "blah"
        assert(hash(obj1) == hash(obj2))

    def test_multikey_hash(self):
        obj1 = DKey("{blah} blee {bloo}", force=MultiDKey)
        obj2 = "{blah} blee {bloo}"
        assert(hash(obj1) == hash(obj2))

    def test_str(self):
        obj1 = DKey[list]("{blah} {bloo}")
        obj2 = "{blah} {bloo}"
        assert(obj1[:] == obj2)
        assert(str(obj1) == obj2)

    def test_getitem_succeeds(self):
        obj1 = DKey[list]("{blah} {bloo}")
        assert(obj1[0,0] == "{blah}")
        assert(obj1[0,1] == "{bloo}")

    def test_get_succeeds(self):
        obj1 = DKey[list]("{blah} {bloo}")
        assert(obj1.get(0,0) == "blah")
        assert(obj1.get(0,1) == "bloo")

class TestNonDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        match DKey("blah", force=NonDKey):
            case NonDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, NonDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x


    def test_basic_by_class_access(self):
        match DKey[False]("blah"):
            case NonDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, NonDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), type(x)


    def test_basic_unannotated(self):
        match DKey("blah"):
            case NonDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, NonDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x

    def test_eq(self):
        obj1 = DKey("blah", force=NonDKey)
        obj2 = DKey("blah", force=NonDKey)
        assert(obj1 == obj2)

    def test_eq_str(self):
        obj1 = DKey("blah", force=NonDKey)
        obj2 = "blah"
        assert(obj1 == obj2)

    def test_eq_not_implemented(self):
        obj1 = DKey("blah", force=NonDKey)
        obj2 = 21
        assert(not (obj1 == obj2))

    def test_hash(self):
        obj1 = DKey("blah", implicit=False, force=NonDKey)
        obj2 = "blah"
        assert(hash(obj1) == hash(obj2))


    def test_format(self):
        obj1 = DKey("blah", implicit=False, force=NonDKey)
        assert(str(obj1) == f"{obj1:w}")
        assert(str(obj1) == f"{obj1:i}")
        assert(str(obj1) == f"{obj1:wi}")


    def test_empty_str(self):
        match DKey(""):
            case NonDKey():
                assert(True)
            case x:
                assert(False), x

class TestIndirectDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    @pytest.mark.parametrize("name", ["blah", "blah_"])
    def test_basic(self, name):
        match DKey(name, implicit=True, force=IndirectDKey):
            case IndirectDKey() as x:
                assert(not hasattr(x, "__dict__"))
                assert(isinstance(x, IndirectDKey))
                assert(isinstance(x, DKey))
                assert(isinstance(x, Strang))
                assert(isinstance(x, str))
            case x:
                assert(False), x

    def test_basic_by_class_access(self):
        match DKey[Mapping]("{blah}"):
            case IndirectDKey():
                assert(True)
            case x:
                assert(False), type(x)

    @pytest.mark.parametrize("name", ["blah", "blah_"])
    def test_eq(self, name):
        obj1 = DKey(name, implicit=True, force=IndirectDKey)
        obj2 = DKey(name, implicit=True, force=IndirectDKey)
        assert(obj1 == obj2)

    @pytest.mark.parametrize("name", ["blah"])
    def test_eq_with_underscore(self, name):
        obj1 = DKey(name, implicit=True, force=IndirectDKey)
        obj2 = DKey(f"{name}_", implicit=True, force=IndirectDKey)
        assert(obj1 == obj2)

    @pytest.mark.parametrize("name", ["blah", "blah_"])
    def test_eq_str(self, name):
        obj1 = DKey(name, implicit=True, force=IndirectDKey)
        obj2 = name
        assert(obj1 == obj2)

    def test_eq_indirect(self):
        obj1 = DKey("blah", force=IndirectDKey, implicit=True)
        obj2 = "blah_"
        assert(obj1 == obj2)

    def test_eq_not_implemented(self):
        obj1 = DKey("blah", force=IndirectDKey, implicit=True)
        obj2 = 21
        assert(isinstance(obj1, IndirectDKey))
        assert(not (obj1 == obj2))

    def test_hash_eq_trimmed(self):
        obj1 = DKey("blah_", implicit=True, force=IndirectDKey)
        obj2 = "blah"
        assert(hash(obj1) == hash(obj2))



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest
from jgdv.structs.strang import Strang, CodeReference

logging = logmod.root

class TestBuildAppropriate:

    def test_sanity(self):
        assert(True is True)

    def test_simple(self):
        obj = Strang("group::tail.a.b.c")
        assert(isinstance(obj, Strang))
        assert(not isinstance(obj, CodeReference))


    @pytest.mark.xfail
    def test_simple_coderef(self):
        obj = Strang("fn::tail.a.b.c:build_fn")
        assert(isinstance(obj, Strang))
        assert(isinstance(obj, CodeReference))

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    import pathlib as pl
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from .._interface import ChainGuard_i

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

try:
    import tomli_w

    class TomlWriter_m:
        """ A mixin for adding toml-writing functionality """

        def __str__(self:ChainGuard_i) -> str:
            return tomli_w.dumps(self._table())

        def to_file(self:ChainGuard_i, path:pl.Path) -> None:
            path.write_text(str(self))

except ImportError:
    logging.debug("No Tomli-w found, ChainGuard will not write toml, only read it")

    class TomlWriter_m: # type: ignore[no-redef]
        """ A fallback mixin for when toml-writing isnt available"""

        def to_file(self, path:pl.Path) -> None:
            msg = "Tomli-w isn't installed, so ChainGuard can't write, only read"
            raise NotImplementedError(msg)



#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from . import _interface as API # noqa: N812
from .annotate import SubAnnotate_m

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class SubRegistry_m(SubAnnotate_m):
    """ Create Subclasses in a registry

    By doing:

    class MyReg(SubRegistry_m):
        _registry : dict[str, type] = {}

    class MyClass(MyReg['blah']: ...

    MyClass is created as a subclass of MyReg, with a parameter set to 'blah'.
    This is added into MyReg._registry
    """
    _registry : ClassVar[dict] = {}

    @classmethod
    def __init_subclass__(cls, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        logging.debug("Registry Subclass: %s : %s : %s", cls, args, kwargs)
        super().__init_subclass__(*args, **kwargs)
        match getattr(cls, "_registry", None):
            case None:
                logging.debug("Creating Registry: %s : %s : %s", cls.__name__, args, kwargs)
                cls._registry = {}
            case _:
                pass
        match cls.cls_annotation():
            case None:
                logging.debug("No Annotation")
                pass
            case x if x in cls._registry and issubclass(cls, (current:=cls._registry[x])):
                logging.debug("Overriding : %s : %s : %s : (%s) : %s", cls.__name__, args, kwargs, x, current)
                cls._registry[x] = cls
            case x if x not in cls._registry:
                logging.debug("Registering: %s : %s : %s : (%s)", cls.__name__, args, kwargs, x)
                cls._registry.setdefault(x, cls)

    @classmethod
    def __class_getitem__(cls:type, *params:Any) -> type: # type:ignore  # noqa: ANN401
        match cls._registry.get(params[0], None):  # type: ignore[attr-defined]
            case None:
                logging.debug("No Registered annotation class: %s :%s", cls, params)
                return super().__class_getitem__(*params)  # type: ignore[misc]
            case x:
                return x

    @classmethod
    def get_registered(cls, *, param:Maybe=None) -> Self:
        param = param or cls.cls_annotation()
        return cls._registry.get(param, cls)

    @classmethod
    def maybe_subclass(cls, *, param:Maybe=None) -> Maybe[Self]:
        param = param or cls.cls_annotation()
        return cls._registry.get(param, None)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ERA001
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

from . import _interface as API # noqa: N812
from .subclasser import Subclasser

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, NewType, _caller  # type: ignore[attr-defined]
from types import GenericAlias
from typing import TypeAliasType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from types import resolve_bases
from pydantic import BaseModel, create_model

if TYPE_CHECKING:
   from jgdv import Maybe, Rx
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Vars

##--| Body

class SubAnnotate_m:
    """
    A Mixin to create simple subclasses through annotation.
    Annotation var name can be customized through the subclass kwarg 'annotate_to'.
    eg:

    class MyExample(SubAnnotate_m, annotate_to='blah'):
        pass

    a_sub = MyExample[int]
    a_sub.__class__.blah == int

    """
    __slots__ = ()

    __builder     : ClassVar[Subclasser]  = Subclasser()
    _annotate_to  : ClassVar[str]         = API.AnnotationTarget

    def __init_subclass__(cls, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        """ On init of a subclass, ensure it's annotation target is set

        """
        super().__init_subclass__(*args)
        match kwargs.pop(API.AnnotateKWD, None):
            case str() as target:
                logging.debug("Annotate Subclassing: %s : %s", cls, kwargs)
                cls._annotate_to = target
                setattr(cls, cls._annotate_to, None)
            case None if not hasattr(cls, cls._annotate_to):
                setattr(cls, cls._annotate_to, None)
            case _:
                pass

    @classmethod
    @ftz.cache
    def __class_getitem__[T:SubAnnotate_m](cls:type[T], *params:Any) -> type[T]:  # noqa: ANN401
        """ Auto-subclass as {cls.__name__}[param]

        Caches results to avoid duplicates
        """
        logging.debug("Annotating: %s : %s : (%s)", cls.__name__, params, cls._annotate_to)  # type: ignore[attr-defined]
        match params:
            case []:
                return cls
            case _:
                return cls.__builder.annotate(cls, *params)

    @classmethod
    def cls_annotation(cls) -> Maybe[str]:
        return getattr(cls, cls._annotate_to, None)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import builtins
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from .param_spec import ParamSpec
from .param_spec.core import ToggleParam, KeyParam, AssignParam, PositionalParam
from .param_spec.defaults import HelpParam, VerboseParam, SeparatorParam

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

def _remap_type(data:dict) -> type:
    """
    Get a specific type of parameter, using provided data

    needs to handle:
    - using separator, prefix and type to select toggle/key/assign/positional
    - TODO using type to select toggle, literal
    - TODO using count to select repeatable
    - TODO using choice to select choice params
    """

    match data:
        case {"separator": str() as x } if bool(x):
            return AssignParam
        case {"prefix" : int() }:
            return PositionalParam
        case {"type": x } if x is not bool:
            return KeyParam
        case _:
            return ToggleParam

##--|

class ParamSpecMaker_m:
    __slots__ = ()

    @staticmethod
    def build_param(**kwargs:Any) -> ParamSpec:
        """ Utility method for easily making paramspecs """
        data = dict(kwargs)
        # Parse the name
        match ParamSpec._processor.parse_name(data['name']):
            case None:
                pass
            case dict() as parsed:
                data.update(parsed)

        # remap to a specific paramspec type
        refined = _remap_type(data)
        # build it, using original kwargs
        return refined(**kwargs)



#!/usr/bin/env python3
"""
TEST File updated

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from ..timeblock_ctx import TimeBlock_ctx, TrackTime

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:
class TestTimeblockCtx:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    ##--|
    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, ANN001, B011, ANN204, ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import warnings
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from .. import _interface as API  # noqa: N812
from .._core import (
    Decorator,
    MonotonicDec,
    IdempotentDec,
    MetaDec,
    DataDec,
)

# ##-- end 1st party imports

class _Utils:

    @pytest.fixture(scope="function")
    def dec(self): # type: ignore
        return Decorator()

    @pytest.fixture(scope="function")
    def mdec(self):

        class MDec(MonotonicDec):
            pass

        return MDec()

    @pytest.fixture(scope="function")
    def idec(self):
        return IdempotentDec()

    @pytest.fixture(scope="function")
    def a_class(self):

        class Basic:

            def simple(self):
                return 2

        return Basic

    @pytest.fixture(scope="function")
    def a_method(self):

        class Basic:

            def simple(self):
                return 2

        return Basic.simple

    @pytest.fixture(scope="function")
    def a_fn(self):

        def simple():
            return 2

        return simple

##--|

class TestDFormDiscrimination(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_is_fn(self, dec, a_fn):
        match dec._discrim_form(a_fn):
            case API.DForm_e.FUNC:
                assert(True)
            case x:
                assert(False), x

    def test_is_instance_method(self, dec, a_class):
        inst = a_class()
        match dec._discrim_form(inst.simple):
            case API.DForm_e.METHOD:
                assert(True)
            case x:
                assert(False), x

    def test_is_method(self, dec, a_method):
        match dec._discrim_form(a_method):
            case API.DForm_e.METHOD:
                assert(True)
            case x:
                assert(False), x

    def test_is_class(self, dec, a_class):
        match dec._discrim_form(a_class):
            case API.DForm_e.CLASS:
                assert(True)
            case x:
                assert(False), x

    def test_instance(self, dec, a_class):
        with pytest.raises(TypeError):
            dec._discrim_form(a_class())

class TestDecorator(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_init(self, dec):
        mark = f"{API.ANNOTATIONS_PREFIX}:{dec.__class__.__name__}"
        data = f"{API.ANNOTATIONS_PREFIX}:{API.DATA_SUFFIX}"
        assert(dec.mark_key() == mark)
        assert(dec.data_key() == data)

    @pytest.mark.parametrize("name", ["blah", "bloo", "blee"])
    def test_custom_prefix(self, name):
        dec = Decorator(prefix=name)
        assert(dec.mark_key() == f"{name}:{dec.__class__.__name__}")
        assert(dec.data_key() == f"{name}:{API.DATA_SUFFIX}")

    @pytest.mark.parametrize("name", ["blah", "bloo", "blee"])
    def test_custom_mark(self, name):
        dec = Decorator(mark=name)
        assert(dec.mark_key() == f"{API.ANNOTATIONS_PREFIX}:{name}")
        assert(dec.data_key() == f"{API.ANNOTATIONS_PREFIX}:{API.DATA_SUFFIX}")

    @pytest.mark.parametrize("name", ["blah", "bloo", "blee"])
    def test_custom_data(self, name):
        dec = Decorator(data=name)
        assert(dec.mark_key() == f"{API.ANNOTATIONS_PREFIX}:{dec.__class__.__name__}")
        assert(dec.data_key() == f"{API.ANNOTATIONS_PREFIX}:{name}")


    def test_decorating_without_instantiation(self):

        class SimpleDec(Decorator):

            def __init__(self, extra=None):
                super().__init__()
                self._extra = extra or []

            def __call__(self, target):
                def simple_wrapped(*args, **kwargs):
                    return [*target(*args, **kwargs), 2, *self._extra]

                return simple_wrapped

        @SimpleDec()
        def test_fn():
            return [1]

        assert(test_fn() == [1,2])


    def test_decorating_with_instantiation(self):

        class SimpleDec(Decorator):

            def __init__(self, extra=None):
                super().__init__()
                self._extra = extra or []

            def __call__(self, target):
                def simple_wrapped(*args, **kwargs):
                    return [*target(*args, **kwargs), 2, *self._extra]

                return simple_wrapped

        @SimpleDec(extra=[3,4,5])
        def test_fn():
            return [1]

        assert(test_fn() == [1, 2, 3, 4, 5])


class TestMarking(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_mark_fn(self, dec, a_fn):
        assert(not dec.is_marked(a_fn))
        dec.apply_mark(a_fn)
        assert(dec.is_marked(a_fn))
        assert(dec._mark_key in a_fn.__annotations__)
        assert(dec._data_key not in a_fn.__annotations__)

    def test_mark_method(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        assert(not dec.is_marked(a_class.simple))
        dec.apply_mark(a_class.simple)
        assert(dec.is_marked(a_class.simple))

    def test_mark_method_doesnt_mark_class(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        assert(not dec.is_marked(a_class.simple))
        dec.apply_mark(a_class.simple)
        assert(dec.is_marked(a_class.simple))
        assert(not dec.is_marked(a_class))

    def test_mark_method_survives_instantiation(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        assert(not dec.is_marked(a_class.simple))
        dec.apply_mark(a_class.simple)
        obj = a_class()
        assert(dec.is_marked(obj.simple))

    def test_mark_method_survives_subclassing(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        assert(not dec.is_marked(a_class.simple))
        dec.apply_mark(a_class.simple)
        assert(dec.is_marked(a_class.simple))

        class BasicSub(a_class):
            pass

        assert(dec.is_marked(BasicSub.simple))

    def test_mark_class(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        dec.apply_mark(a_class)
        assert(dec.is_marked(a_class))

    def test_mark_class_survives_instantiation(self, a_class, dec):
        assert(not dec.is_marked(a_class))
        dec.apply_mark(a_class)
        assert(dec.is_marked(a_class))
        obj = a_class()
        assert(dec.is_marked(obj))

class TestAnnotation(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_fn_is_not_annotated(self, dec, a_fn):
        assert(not dec.is_annotated(a_fn))

    def test_method_is_not_annotated_(self, dec, a_method):
        assert(not dec.is_annotated(a_method))

    def test_class_is_not_annotated(self, dec, a_class):
        assert(not dec.is_annotated(a_class))

    def test_instance_is_not_annotated(self, dec, a_class):
        obj = a_class()
        assert(not dec.is_annotated(obj))

class TestWrapping(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_fn_unwrap_default(self, mdec, a_fn):
        assert(not mdec.is_marked(a_fn))
        unwrapped = mdec._unwrap(a_fn)
        assert(unwrapped is a_fn)

    def test_method_unwrap_default(self, mdec, a_method):
        assert(not mdec.is_marked(a_method))
        unwrapped = mdec._unwrap(a_method)
        assert(unwrapped is a_method)

    def test_class_unwrap_default(self, mdec, a_class):
        assert(not mdec.is_marked(a_class))
        unwrapped = mdec._unwrap(a_class)
        assert(unwrapped is a_class)

    def test_fn_wrap(self, mdec, a_fn):
        assert(not mdec.is_marked(a_fn))
        decorated = mdec(a_fn)
        assert(decorated is not a_fn)

    def test_fn_unwrap(self, mdec, a_fn):
        assert(not mdec.is_marked(a_fn))
        decorated = mdec(a_fn)
        unwrapped = mdec._unwrap(decorated)
        assert(unwrapped is a_fn)
        assert(unwrapped is not decorated)

    def test_method_wrap(self, mdec, a_method):
        assert(not mdec.is_marked(a_method))
        decorated = mdec(a_method)
        assert(decorated is not a_method)
        assert(mdec.is_marked(a_method))
        assert(mdec.is_marked(decorated))

    def test_method_unwrap(self, mdec, a_method):
        assert(not mdec.is_marked(a_method))
        decorated = mdec(a_method)
        unwrapped = mdec._unwrap(decorated)
        assert(unwrapped is a_method)
        assert(unwrapped is not decorated)

    def test_class_wrap(self, mdec, a_class):
        assert(not mdec.is_marked(a_class))
        decorated = mdec(a_class)
        assert(decorated is not a_class)
        assert(mdec.is_marked(a_class))
        assert(mdec.is_marked(decorated))

    def test_class_unwrap(self, mdec, a_class):
        assert(not mdec.is_marked(a_class))
        decorated = mdec(a_class)
        unwrapped = mdec._unwrap(decorated)
        assert(unwrapped is not a_class)
        assert(unwrapped is decorated)

    def test_unwrap_depth_simple(self, dec):

        def simple():
            return 2

        assert(dec._unwrapped_depth(simple) == 0)
        w1 = ftz.update_wrapper(lambda fn: fn(), simple)
        assert(dec._unwrapped_depth(w1) == 1)
        w2 = ftz.update_wrapper(lambda fn: fn(), w1)
        assert(dec._unwrapped_depth(w2) == 2)
        w3 = ftz.update_wrapper(lambda fn: fn(), w2)
        assert(dec._unwrapped_depth(w3) == 3)

##--|

class TestIdempotent(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_mark_of_class_persists_to_instances(self, idec):
        class Basic:

            @idec
            def simple(self):
                pass

        instance = Basic()
        assert(idec.is_marked(Basic.simple))
        assert(idec.is_marked(instance.simple))
        assert(not idec.is_annotated(Basic.simple))
        assert(not idec.is_annotated(instance.simple))

    def test_fn_wrap_idempotent(self, idec, a_fn):
        assert(not idec.is_marked(a_fn))
        d1 = idec(a_fn)
        d2 = idec(d1)
        assert(d1 is not a_fn)
        assert(d2 is not a_fn)
        assert(d2 is d1)
        assert(idec.is_marked(a_fn))
        assert(idec.is_marked(d1))
        assert(idec.is_marked(d2))
        assert(idec._unwrapped_depth(d1) == idec._unwrapped_depth(d2))

    def test_doesnt_annotate(self, idec):
        assert(not bool(idec._build_annotations_h(None, [])))

class TestMonotonic(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_mark_of_class_persists_to_instances(self, mdec):

        class Basic:

            @mdec
            def simple(self):
                pass

        instance = Basic()
        assert(mdec.is_marked(Basic.simple))
        assert(mdec.is_marked(instance.simple))
        assert(mdec._mark_key in Basic.simple.__annotations__)

    def test_fn_retains_correct_type(self):

        class Dec1(MonotonicDec):
            pass

        class Dec2(MonotonicDec):
            pass

        @Dec1()
        @Dec2()
        def testfn():
            pass

        match Dec1()._discrim_form(testfn):
            case API.DForm_e.FUNC:
                assert(True)
            case x:
                assert(False), x

    def test_method_retains_correct_type(self):

        class Dec1(MonotonicDec):
            pass

        class Dec2(MonotonicDec):
            pass

        class TestClass:

            @Dec1()
            @Dec2()
            def testfn(self):
                pass

        match Dec1()._discrim_form(TestClass.testfn):
            case API.DForm_e.METHOD:
                assert(True)
            case x:
                assert(False), x

class TestMetaDecorator(_Utils):

    @pytest.fixture(scope="function")
    def a_meta_dec(self):
        return MetaDec("example")

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_init(self, a_meta_dec, dec):
        assert(isinstance(a_meta_dec, Decorator))
        assert(issubclass(a_meta_dec.__class__, Decorator))
        assert(a_meta_dec._data == ["example"])

    def test_basic_wrap_fn(self, a_meta_dec, a_fn):
        assert(not a_meta_dec.is_annotated(a_fn))
        wrapped = a_meta_dec(a_fn)
        assert(wrapped is a_fn)
        assert(a_meta_dec.get_annotations(wrapped) == ["example"])

class TestDataDecorator(_Utils):

    @pytest.fixture(scope="function")
    def ddec(self):
        return DataDec("aval")

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self, ddec, a_fn):
        wrapped = ddec(a_fn)
        assert(bool(ddec._build_annotations_h(None, [])))
        assert(ddec.get_annotations(wrapped) == ["aval"])

class TestClassDecoration(_Utils):

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_add_new_method(self):
        """ Modify the decorated class"""
        class ExDecorator(IdempotentDec):

            def bmethod(self, val):
                return val + self._val

            def _wrap_class_h(self, target:type):
                # Gets the unbound method and binds it to the target
                setattr(target, "bmethod", self.__class__.bmethod) # noqa: B010


        @ExDecorator()
        class Basic:

            def __init__(self, val=None):
                self._val = val or 2

            def amethod(self):
                return 2

        inst = Basic()
        assert(inst.amethod() == 2)
        assert(inst.bmethod(2) == 4)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: PLR0912
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from collections.abc import Callable, Generator, Iterable, Iterator, Mapping, MutableMapping, Sequence
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Maybe
from jgdv.enums import LoopControl_e
from jgdv.structs.dkey import DKey
# ##-- end 1st party imports

from jgdv import identity_fn

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

MARKER       : Final[str]        = ".marker"
walk_ignores : Final[list[str]]  = ['.git', '.DS_Store', "__pycache__"] # TODO use a .ignore file
walk_halts   : Final[list[str]]  = [".doot_ignore"]

class PathManip_m:
    """
      A Mixin for common path manipulations
    """

    def _calc_path_parts(self, fpath:pl.Path, roots:list[pl.Path]) -> dict:
        """ take a path, and get a dict of bits which aren't methods of Path
          if no roots are provided use cwd
        """
        assert(fpath is not None)
        assert(isinstance(roots, list))

        temp_stem  = fpath
        # This handles "a/b/c.tar.gz"
        while temp_stem.stem != temp_stem.with_suffix("").stem:
            temp_stem = temp_stem.with_suffix("")

        return {
            'rpath'   : self._get_relative(fpath, roots),
            'fstem'   : temp_stem.stem,
            'fparent' : fpath.parent,
            'fname'   : fpath.name,
            'fext'    : fpath.suffix,
            'pstem'   : fpath.parent.stem,
            }

    def _build_roots(self, *sources:Mapping, roots:Maybe[list[str|DKey]]=None) -> list[pl.Path]:
        """
        convert roots from keys to paths
        """
        roots   = roots or []
        results = []
        for root in roots:
            root_key = DKey(root, fallback=root, mark=DKey.Mark.PATH)
            results.append(root_key.expand(*sources))
        else:
            return results

    def _get_relative(self, fpath:pl.Path, roots:Maybe[list[pl.Path]]=None) -> pl.Path:
        """ Get relative path of fpath.
          if no roots are provided, default to using cwd
        """
        logging.debug("Finding Relative Path of: %s using %s", fpath, roots)
        if not fpath.is_absolute():
            return fpath

        roots = roots or [pl.Path.cwd()]

        for root_path in roots:
            try:
                return fpath.relative_to(root_path)
            except ValueError:
                continue

        msg = f"{fpath} is not able to be made relative"
        raise ValueError(msg, roots)

    def _shadow_path(self, rpath:pl.Path, shadow_root:pl.Path) -> pl.Path:
        """ take a relative path, apply it onto a root to create a shadowed location """
        raise NotImplementedError()

    def _find_parent_marker(self, fpath:pl.Path, marker:Maybe[str]=None) -> Maybe[pl.Path]:
        """ Go up the parent list to find a marker file, return the dir its in """
        marker = marker or MARKER
        for p in fpath.parents:
            if (p / marker).exists():
                return p

        return None

    def _normalize(self, path:pl.Path, *, root:Maybe[pl.Path]=None, symlinks:bool=False) -> pl.Path:
        """
          a basic path normalization
          expands user, and resolves the location to be absolute
        """
        result : pl.Path = path
        if symlinks and path.is_symlink():
            msg = "symlink normalization"
            raise NotImplementedError(msg, path)

        match result.parts:
            case ["~", *_]:
                result = result.expanduser().resolve()
            case ["/", *_]:
                pass
            case _ if root:
                result = (root / path).expanduser().resolve()
            case _:
                pass

        return result

class Walker_m:
    """ A Mixin for walking directories,
      written for py<3.12
      """
    control_e = LoopControl_e

    def walk_all(self, roots:list[pl.Path], *, exts:Maybe[list[str]]=None, rec:bool=False, fn:Maybe[Callable]=None) -> list[dict]:
        """
        walk all available targets,
        and generate unique names for them
        """
        result : list = []
        exts          = exts or []
        match rec:
            case True:
                for root in roots:
                    result += self.walk_target_deep(root, exts=exts, fn=fn)
            case False:
                for root in roots:
                    result += self.walk_target_shallow(root, exts=exts, fn=fn)

        return result

    def walk_target_deep(self, target:pl.Path, *, exts:Maybe[list[str]]=None, fn:Maybe[Callable]=None) -> Generator[pl.Path]:
        logging.info("Deep Walking Target: %s : exts=%s", target, exts)
        exts = exts or []
        fn   = fn or identity_fn
        if not target.exists():
            return None

        queue = [target]
        while bool(queue):
            current = queue.pop()
            if not current.exists():
                continue
            if current.name in walk_ignores:
                continue
            if current.is_dir() and any((current / x).exists() for x in walk_halts):
                continue
            if bool(exts) and current.is_file() and current.suffix not in exts:
                continue
            match fn(current):
                case self.control_e.yes:
                    yield current
                case True if current.is_dir():
                    queue += sorted(current.iterdir())
                case True | self.control_e.yesAnd:
                    yield current
                    if current.is_dir():
                        queue += sorted(current.iterdir())
                case False | self.control_e.noBut if current.is_dir():
                    queue += sorted(current.iterdir())
                case None | False:
                    continue
                case self.control_e.no | self.control_e.noBut:
                    continue
                case _ as x:
                    msg = "Unexpected filter value"
                    raise TypeError(msg, x)

    def walk_target_shallow(self, target:pl.Path, *, exts:Maybe[list[str]]=None, fn:Maybe[Callable]=None) -> Generator:
        logging.debug("Shallow Walking Target: %s", target)
        exts = exts or []
        fn = fn or identity_fn
        if target.is_file():
            fn_fail = fn(target) in [None, False, self.control_e.no, self.control_e.noBut]
            ignore  = target.name in walk_ignores
            bad_ext = (bool(exts) and target.suffix in exts)
            if not (fn_fail or ignore or bad_ext):
                yield target
            return None

        for x in target.iterdir():
            fn_fail = fn(x) in [None, False, self.control_e.no, self.control_e.noBut]
            ignore  = x.name in walk_ignores
            bad_ext = bool(exts) and x.is_file() and x.suffix not in exts
            if not (fn_fail or ignore or bad_ext):
                yield x



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ERA001
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import sys
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

from jgdv.debugging import TraceBuilder
from ._core import IdempotentDec, MetaDec, MonotonicDec

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Concatenate
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from . import _interface as API # noqa: N812

    from jgdv import Maybe, Either, Method, Func
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard, ParamSpec
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    type Logger = logmod.Logger

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class NoSideEffects(MetaDec):
    """ TODO Mark a Target as not modifying external variables """
    pass

class CanRaise(MetaDec):
    """ TODO mark a target as able to raise certain errors.
    Non-exaustive, doesn't change runtime behaviour,
    just to simplify documentation

    """
    pass

class Breakpoint(IdempotentDec):
    """
      Decorator to attach a breakpoint to a function, without pausing execution
    """

    def __call__[**I, O](self, target:Callable[I,O]) -> Callable[I,O]:
        msg = "needs RunningDebugger"
        raise NotImplementedError(msg)
    # # TODO handle repeats
    # if args[0].breakpoint:

        #     f_code = f.__code__
        #     db = RunningDebugger()
        #     # Ensure trace function is set
        #     sys.settrace(db.trace_dispatch)
        #     if not db.get_break(f_code.co_filename, f_code.co_firstlineno+2):
        #         db.set_break(f_code.co_filename,
        #                     f_code.co_firstlineno+2,
        #                     True)
        #     else:
        #         bp = Breakpoint.bplist[f_code.co_filename,
        #                             f_code.co_firstlineno+2][0]
        #         bp.enable()

        # return self._func(self, *args, **kwargs)

class DoMaybe(MonotonicDec):
    """ Make a fn or method propagate None's """
    type MethMb[Obj,X,**I,O]  = Callable[Concatenate[Obj, X, I],O]
    type FuncMb[X,**I,O]      = Callable[Concatenate[X, I],O]
                                                                                                                     # def __call__[**I, X,Obj, O](self, target:MethMb[Obj,X,I,O]|FuncMb[X,I,O], *args:Any, **kwargs:Any) -> MethMb[Obj,Maybe[X],I,O]|FuncMb[Maybe[X],I,O]: # type: ignore[override]

    @overload # type: ignore[override]
    def __call__[Obj,**I,X,O](self, target:MethMb[Obj,X,I,O], *args:Any, **kwargs:Any) -> MethMb[Obj,Maybe[X],I,O]:  # noqa: ANN401
        pass

    @overload
    def __call__[**I,X,O](self, target:FuncMb[X,I,O], *args:Any, **kwargs:Any) -> FuncMb[Maybe[X],I,O]: # noqa: ANN401
        pass

    def __call__(self, target, *args, **kwargs):
        return MonotonicDec.__call__(self, target, *args, **kwargs)

    @override
    def _wrap_method_h[Obj, X, **I, O](self, meth:MethMb[Obj,X,I,Maybe[O]]) -> MethMb[Obj,Maybe[X],I,Maybe[O]]: # type: ignore[override]

        def _prop_maybe(_self:Obj, fst:Maybe[X], *args:I.args, **kwargs:I.kwargs) -> Maybe[O]:
            match fst:
                case None:
                    return None
                case x:
                    return meth(_self, x, *args, **kwargs)

        return _prop_maybe

    @override
    def _wrap_fn_h[X, **I, O](self, fn:FuncMb[X, I, Maybe[O]]) -> FuncMb[Maybe[X], I, Maybe[O]]: # type: ignore[override]

        def _prop_maybe(fst:Maybe[X], *args:I.args, **kwargs:I.kwargs) -> Maybe[O]:
            match fst:
                case None:
                    return None
                case x:
                    try:
                        return fn(x, *args, **kwargs)
                    except Exception as err:
                        err.with_traceback(TraceBuilder[2:]) # type: ignore[misc]
                        raise

        return _prop_maybe

class DoEither(MonotonicDec):
    """ Either do the fn/method, or propagate the error """

    def _wrap_method_h[X,Y, **I, O, E:Exception](self, meth:Method[Concatenate[X, Y, I], Either[O,E]]) -> API.Decorated[Concatenate[X, Y|E, I], Either[O, E]]: # type: ignore[override]

        def _prop_either(_self:X, fst:Y|E, *args:I.args, **kwargs:I.kwargs) -> Either[O, E]:
            match fst:
                case Exception() as err:
                    return cast("E", err)
                case x:
                    try:
                        return meth(_self, x, *args, **kwargs)
                    except Exception as err:  # noqa: BLE001
                        err.with_traceback(TraceBuilder[2:]) # type: ignore[misc]
                        return cast("E", err)

        return _prop_either

    def _wrap_fn_h[X, **I, O, E:Exception](self, fn:Func[Concatenate[X, I], O]) -> Func[Concatenate[X|E, I], Either[O, E]]: # type: ignore[override]

        def _prop_either(fst:X|E, *args:I.args, **kwargs:I.kwargs) -> Either[O, E]:
            match fst:
                case Exception() as err:
                    return cast("E", err)
                case x:
                    try:
                        return fn(x, *args, **kwargs)
                    except Exception as err:  # noqa: BLE001
                        err.with_traceback(TraceBuilder[2:]) # type: ignore[misc]
                        return cast("E", err)

        return _prop_either



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from .param_spec import ParamSpec
from .core import AssignParam, ToggleParam, KeyParam, LiteralParam

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class RepeatToggleParam(ToggleParam):
    """ TODO A repeatable toggle
    eg: -verbose -verbose -verbose
    """
    desc : str = "A Repeat Toggle"
    pass

class WildcardParam(AssignParam):
    """ TODO a wildcard param that matches any --{key}={val} """

    desc : str = "A Wildcard"

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401, ARG002
        kwargs.setdefault("type", str)
        kwargs['name'] = "*"
        super().__init__(**kwargs)

    def matches_head(self, val:str) -> bool:
        assert(isinstance(self.separator, str))
        match self.prefix:
            case str() as p:
                return (val.startswith(p)
                        and self.separator in val)
            case _:
                return False

    def next_value(self, args:list) -> tuple[str, list, int]:
        logging.debug("Getting Wildcard Key Assingment: %s", args)
        assert(self.separator in args[0]), (self.separator, args[0])
        key,val = self._processor.split_assignment(self, args[0])
        match self.prefix:
            case str():
                return key.removeprefix(self.prefix), [val], 1
            case int():
                return key, [val], 1


class ImplicitParam(ParamSpec):
    """
    A Parameter that is implicit, so doesn't give a help description unless
    forced to
    """
    desc : str = "An Implicit Parameter"

    def help_str(self, *, force:bool=False) -> str:
        return ""

class RepeatableParam(KeyParam):
    """ TODO a repeatable key param
    -key val -key val2 -key val3
    """

    type_ : type
    desc  : str = "A Repeatable Key"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.type_ = list

    def next_value(self, args:list) -> tuple[str, list, int]:
        """ Get as many values as match
        eg: args[-test, 2, -test, 3, -test, 5, -nottest, 6]
        ->  [2,3,5], [-nottest, 6]
        """
        logging.debug("Getting until no more matches: %s : %s", self.name, args)
        assert(self.repeatable)
        result, consumed, remaining  = [], 0, args[:]
        while bool(remaining):
            head, val, *rest = remaining
            if not self.matches_head(head):
                break
            else:
                result.append(val)
                remaining = rest
                consumed += 2

        return self.name, result, consumed

class ChoiceParam(LiteralParam):
    """ TODO A param that must be from a choice of literals
    eg: ChoiceParam([blah, bloo, blee]) : blah | bloo | blee

    """

    desc : str = "A Choice"

    def __init__(self, name, choices:list[str], **kwargs) -> None:
        super().__init__(name=name, **kwargs)
        self._choices = choices

    def matches_head(self, val) -> bool:
        """ test to see if a cli argument matches this param

        Matchs {self.prefix}{self.name} if not an assignment
        Matches {self.prefix}{self.name}{separator} if an assignment
        """
        return val in self._choices

class EntryParam(LiteralParam):
    """ TODO a parameter that if it matches,
    returns list of more params to parse
    """
    desc : str = "An expandable Param"
    pass

class ConstrainedParam(ParamSpec):
    """
    TODO a type of parameter which is constrained in the values it can take, beyond just type.

    eg: {name:amount, constraints={min=0, max=10}}
    """
    constraints  : list[Any]
    desc         : str  = "A Constrained Param"



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv.mixins.annotate import Subclasser

# ##-- end 1st party imports

from . import _interface as API # noqa: N812
from ._core import Decorator, IdempotentDec, MonotonicDec, _DecAnnotate_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, TypeAliasType, _GenericAlias # type: ignore[attr-defined]
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from types import resolve_bases, FunctionType, MethodType, MethodDescriptorType, WrapperDescriptorType, ClassMethodDescriptorType, BuiltinFunctionType, BuiltinMethodType

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    from ._interface import Decorable, Decorated, DForm_e

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Global Vars:
PROTO_SUFFIX : Final[str] = "protocols"
ABSMETHS     : Final[str] = "__abstractmethods__"
IS_ABS       : Final[str] = "__isabstractmethod__"
NAME_MOD     : Final[str] = "P"
##--| Funcs

##--| Body

class CheckProtocols(_DecAnnotate_m):
    """ A Class Decorator to ensure a class has no abc.abstractmethod's
    or unimplemented protocol members

    pass additional protocols when making the decorator, eg::

        @CheckProtocol(Proto1_p, Proto2_p, AbsClass...)
        class MyClass:
        pass

    """
    _annotation_prefix  : ClassVar[str]  = API.ANNOTATIONS_PREFIX
    _data_suffix        : ClassVar[str]   = PROTO_SUFFIX
    _data_key = None

    def get_protos(self, target:type) -> set[Protocol]:
        """ Get the protocols of a type from its mro and annotations """
        # From MRO
        protos = []
        for x in target.mro():
            match x:
                case _ if x in [Protocol,Generic, target, object]:
                    pass
                case TypeAliasType() if issubclass(x.__value__, Protocol):
                    protos.append(x)
                case x if issubclass(x, Protocol):
                    protos.append(x)
                case x if issubclass(x, abc.ABC):
                    protos.append(x)
                case _:
                    pass
        else:
            protos += target.__annotations__.get(self.data_key(), [])
            return set(protos)

    def test_method(self, cls:type, name:str) -> bool:
        """ return True if the named method is abstract still """
        if name == ABSMETHS:
            return False
        match getattr(cls, name, None):
            case None:
                return True
            case FunctionType() as x if hasattr(x, IS_ABS):
                return x.__isabstractmethod__
            case FunctionType() | property():
                return False

    def test_protocol(self, proto:Protocol, cls) -> list[str]:
        """ Returns a list of methods which are defined in the protocol,
        and nowhere else in the mro.

        ie: they are unimplemented protocol requirements

        Can handle type aliases, so long as they actually point to a protocol.
        | eg: type proto_alias = MyProtocol_p
        | where issubclass(MyProtocol_p, Protocol)
        """
        members : set
        result  : list = []
        # Get the members of the protocol/abc
        match proto:
            case type() if issubclass(proto, Protocol):
                non_callable = getattr(proto, "__non_callable_proto_members__", set())
                fields       = getattr(proto, "__annotations__", {})
                non_attrs    = {x for x in proto.__protocol_attrs__ if getattr(proto, x, None) is None}
                members      = set(proto.__protocol_attrs__) - non_callable - fields.keys() - non_attrs
                qualname     = proto.__qualname__
            case type() if issubclass(proto, abc.ABC):
                return []
            case _:
                raise TypeError("Checking a protocol... but it isnt' a protocol", proto)

        # then filter out the implemented ones
        for member in members:
            match getattr(cls, member, None):
                case property():
                    pass
                case None:
                    result.append(member)
                case FunctionType() as meth if qualname in meth.__qualname__:
                    # (as a class, the method isn't actually bound as a method yet, its still a function)
                    result.append(member)
                case FunctionType():
                    pass
                case MethodType() as meth  if qualname in meth.__func__.__qualname__:
                    result.append(member)
                case MethodType() | ftz.cached_property():
                    pass
                case BuiltinFunctionType() | BuiltinMethodType():
                    pass
                case MethodDescriptorType() | WrapperDescriptorType() | ClassMethodDescriptorType():
                    pass
                case _:
                    pass
        else:
            return result

    def validate_protocols(self, cls:type, *, protos:Maybe[list[Protocol]]) -> type:
        still_abstract = set()
        for meth in getattr(cls, ABSMETHS, []):
            if self.test_method(cls, meth):
                still_abstract.add(meth)
        ##--|
        for proto in self.get_protos(cls):
            still_abstract.update(self.test_protocol(proto, cls))
            ##--|
        for proto in protos or []:
            still_abstract.update(self.test_protocol(proto, cls))
            ##--|
        if not bool(still_abstract):
            return cls

        raise NotImplementedError("Class has Abstract Methods",
                                  cls.__qualname__,
                                  f"module:{cls.__module__}",
                                  still_abstract)

class Proto(MonotonicDec):
    """ Decorator to explicitly annotate a class as an implementer of a set of protocols.

    Protocols are annotated into cls._jgdv_protos : set[Protocol]::

        class ClsName(Supers*, P1, P1..., **kwargs):...

    becomes::

        @Protocols(P1, P2,...)
        class ClsName(Supers): ...

    Protocol *definition* remains the usual way::

        class Proto1(Protocol): ...

        class ExtProto(Proto1, Protocol): ...

    """
    needs_args = True
    _checker   : ClassVar[CheckProtocols] = CheckProtocols()
    _protos    : list
    _name_mod  : str
    _mod_mro   : bool
    _check     : bool

    def __init__(self, *protos:Protocol, check:bool=True, mod_mro:bool=False):
        super().__init__(data=PROTO_SUFFIX)
        self._protos    = []
        self._name_mod  = NAME_MOD
        self._mod_mro   = mod_mro
        self._check     = check

        for x in protos:
            match x:
                case TypeAliasType() if isinstance(x.__value__, _GenericAlias):
                    x = x.__value__.__origin__
                case TypeAliasType():
                    x = x.__value__
                case _GenericAlias():
                    x = x.__origin__
                case _:
                    pass

            match x:
                case _ if issubclass(x, Protocol):
                    self._protos.append(x)
                case x:
                     raise TypeError("Tried to attach a non-protocol to a class", x)
        else:
            pass

    def _validate_target_h(self, target:Decorable, form:DForm_e, args:Maybe[list]=None) -> None:
        match target:
            case type() if issubclass(target, Protocol):
                raise TypeError("Don't use @Proto to combine protocols, use normal inheritance", target)
            case type():
                pass
            case _:
                raise TypeError("Unexpected type passed for protocol annotation")

    def _wrap_class_h(self, cls:type) -> Maybe[type]:
        """ Logic for inserting the protocol into the given cls. """
        new_name = Subclasser.decorate_name(cls, self._name_mod)
        protos = self._checker.get_protos(cls)
        protos.update(self._protos)

        try:
            if self._mod_mro:
                modified = self._build_mro(cls)
                customized = self._builder.make_subclass(new_name, cls, mro=modified)
            else:
                customized = cls
        except TypeError as err:
            raise TypeError(*err.args, cls, self._protos, protos) from None

        self.annotate_decorable(customized)
        match self._check:
            case True:
                self._checker.validate_protocols(customized, protos=self._protos)
            case _:
                pass
        ##--|
        return customized

    def _build_mro(self, cls) -> list:
        import typing
        match cls.mro():
            case [*xs, typing.Protocol, x] if x is object:
                base = [*xs, *self._protos, object]
            case [*xs, x] if x is object:
                base = [*xs, *self._protos, object]
            case x:
                raise TypeError(type(x))

        return base

    def _build_annotations_h(self, target:Decorable, current:list) -> Maybe[list]:
        updated = current[:]
        updated += [x for x in self._protos if x not in current]
        updated += self._checker.get_protos(target)
        return updated

    @staticmethod
    def get(cls:type) -> list[Protocol]:
        """ Get a List of protocols the class is annotated with """
        return list(Proto._checker.get_protos(cls))

    @staticmethod
    def validate_protocols(cls:type, *protos:Protocol):
        return Proto._checker.validate_protocols(cls, protos=protos)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from jgdv._abstract.stdlib_protos import Mapping_p

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    import pathlib as pl
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
type TomlTypes = (str | int | float | bool | list['TomlTypes']
                  | dict[str,'TomlTypes'] | datetime.datetime)
type ProxyWrapper[T] = Callable[[TomlTypes], T]
# Body:

class ChainProxy_p(Protocol):
    """ The proxy interface

    Used for special access like::

        cg.on_fail(...).val()

    """

    def __call__[T](self, wrapper:Maybe[ProxyWrapper[T]]=None, fallback_wrapper:Maybe[ProxyWrapper[T]]=None) -> T: ...

    def __getattr__(self, attr:str) -> Self: ...

    def __getitem__(self, keys:str|tuple[str]) -> Self: ...

class ProxyEntry_p(Protocol):

    def on_fail(self, fallback:Any, types:Maybe[Any]=None, *, non_root:bool=False) -> ChainProxy_p: ...  # noqa: ANN401

    def first_of(self, fallback:Any, types:Maybe[Any]=None) -> ChainProxy_p: ...  # noqa: ANN401

    def all_of(self, fallback:Any, types:Maybe[Any]=None) -> ChainProxy_p: ...  # noqa: ANN401

    def flatten_on(self, fallback:Any) -> ChainProxy_p: ...  # noqa: ANN401

    def match_on(self, **kwargs:tuple[str,Any]) -> ChainProxy_p: ...

@runtime_checkable
class ChainGuard_p(ProxyEntry_p, Mapping_p, Protocol):
    """ The interface for a base ChainGuard object """

    def get(self, key:str, default:Maybe[TomlTypes]=None) -> Maybe[TomlTypes]: ...

    @classmethod
    def read[T:ChainGuard_p](cls:T, text:str) -> T: ...

    @classmethod
    def from_dict(cls, data:dict[str, TomlTypes]) -> Self: ...

    @classmethod
    def load(cls, *paths:str|pl.Path) -> Self: ...

    @classmethod
    def load_dir(cls, dirp:str|pl.Path) -> Self: ...


    @staticmethod
    def report_defaulted() -> list[str]: ...

    def to_file(self, path:pl.Path) -> None: ...

    def _table(self) -> dict[str,TomlTypes]: ...

class ChainGuard_i(ChainGuard_p, Protocol):
    pass



#!/usr/bin/env python3
"""
TEST File updated

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

##--|
from .. import _interface as API # noqa: N812

from ..debug import StrangDebug
from ..strang import Strang
##--|

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol
# from . import _interface as API # noqa: N812
# Dataclasses:
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

#
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

@pytest.mark.skip
class TestStrangDebug:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        result = StrangDebug.debug(Strang, "head::body")
        logging.debug(result)
        assert(False)

    def test_with_args(self):
        result = StrangDebug.debug(Strang, "head::body[test]")
        logging.debug(result)
        assert(False)

    def test_with_multi_args(self):
        result = StrangDebug.debug(Strang, "head::body[test,blah]")
        logging.debug(result)
        assert(False)

    def test_uuid(self):
        result = StrangDebug.debug(Strang, "head::body.<uuid>")
        logging.debug(result)
        assert(False)

    def test_mark(self):
        result = StrangDebug.debug(Strang, "head::body.$head$")
        logging.debug(result)
        assert(False)

    def test_type_val(self):
        result = StrangDebug.debug(Strang, "head::body.<int:10>")
        logging.debug(result)
        assert(False)

    def test_complex(self):
        result = StrangDebug.debug(Strang, "head::body.<int:10>.d.<int:4>.$head$[<uuid>]")
        logging.debug(result)
        assert(False)

